"""
Builds out and synchronizes yum repo mirrors.
Initial support for rsync, perhaps reposync coming later.

Copyright 2006-2007, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""
import logging
import os
import os.path
import pipes
import stat
import shutil
from typing import Optional, Union

from cobbler import utils
from cobbler import download_manager
from cobbler.enums import RepoArchs, RepoBreeds, MirrorType
from cobbler.utils import os_release

HAS_LIBREPO = True
try:
    import librepo
except:
    HAS_LIBREPO = False


def repo_walker(top, func, arg):
    """
    Directory tree walk with callback function.

    For each directory in the directory tree rooted at top (including top itself, but excluding '.' and '..'), call
    func(arg, dirname, fnames). dirname is the name of the directory, and fnames a list of the names of the files and
    subdirectories in dirname (excluding '.' and '..').  func may modify the fnames list in-place (e.g. via del or slice
    assignment), and walk will only recurse into the subdirectories whose names remain in fnames; this can be used to
    implement a filter, or to impose a specific order of visiting.  No semantics are defined for, or required of, arg,
    beyond that arg is always passed to func. It can be used, e.g., to pass a filename pattern, or a mutable object
    designed to accumulate statistics. Passing None for arg is common.

    :param top: The directory that should be taken as root. The root dir will also be included in the processing.
    :param func: The function that should be executed.
    :param arg: The arguments for that function.
    """
    try:
        names = os.listdir(top)
    except os.error:
        return
    func(arg, top, names)
    for name in names:
        name = os.path.join(top, name)
        try:
            st = os.lstat(name)
        except os.error:
            continue
        if stat.S_ISDIR(st.st_mode):
            repo_walker(name, func, arg)


class RepoSync:
    """
    Handles conversion of internal state to the tftpboot tree layout.
    """

    # ==================================================================================

    def __init__(self, collection_mgr, tries: int = 1, nofail: bool = False):
        """
        Constructor

        :param collection_mgr: The object which holds all information in Cobbler.
        :param tries: The number of tries before the operation fails.
        :param nofail: This sets the strictness of the reposync result handling.
        :param logger: The logger to audit all actions with.
        """
        self.verbose = True
        self.api = collection_mgr.api
        self.collection_mgr = collection_mgr
        self.distros = collection_mgr.distros()
        self.profiles = collection_mgr.profiles()
        self.systems = collection_mgr.systems()
        self.settings = collection_mgr.settings()
        self.repos = collection_mgr.repos()
        self.rflags = self.settings.reposync_flags
        self.tries = tries
        self.nofail = nofail
        self.logger = logging.getLogger()
        self.dlmgr = download_manager.DownloadManager(self.collection_mgr)

        self.logger.info("hello, reposync")

    # ===================================================================

    def run(self, name: Optional[str] = None, verbose: bool = True):
        """
        Syncs the current repo configuration file with the filesystem.

        :param name: The name of the repository to synchronize.
        :param verbose: If the action should be logged verbose or not.
        """

        self.logger.info("run, reposync, run!")

        try:
            self.tries = int(self.tries)
        except:
            utils.die("retry value must be an integer")

        self.verbose = verbose

        report_failure = False
        for repo in self.repos:
            if name is not None and repo.name != name:
                # Invoked to sync only a specific repo, this is not the one
                continue
            elif name is None and not repo.keep_updated:
                # Invoked to run against all repos, but this one is off
                self.logger.info("%s is set to not be updated" % repo.name)
                continue

            repo_mirror = os.path.join(self.settings.webdir, "repo_mirror")
            repo_path = os.path.join(repo_mirror, repo.name)

            if not os.path.isdir(repo_path) and not repo.mirror.lower().startswith("rhn://"):
                os.makedirs(repo_path)

            # Set the environment keys specified for this repo and save the old one if they modify an existing variable.

            env = repo.environment
            old_env = {}

            for k in list(env.keys()):
                self.logger.debug("setting repo environment: %s=%s" % (k, env[k]))
                if env[k] is not None:
                    if os.getenv(k):
                        old_env[k] = os.getenv(k)
                    else:
                        os.environ[k] = env[k]

            # Which may actually NOT reposync if the repo is set to not mirror locally but that's a technicality.

            for x in range(self.tries + 1, 1, -1):
                success = False
                try:
                    self.sync(repo)
                    success = True
                    break
                except:
                    utils.log_exc()
                    self.logger.warning("reposync failed, tries left: %s" % (x - 2))

            # Cleanup/restore any environment variables that were added or changed above.

            for k in list(env.keys()):
                if env[k] is not None:
                    if k in old_env:
                        self.logger.debug("resetting repo environment: %s=%s" % (k, old_env[k]))
                        os.environ[k] = old_env[k]
                    else:
                        self.logger.debug("removing repo environment: %s=%s" % (k, env[k]))
                        del os.environ[k]

            if not success:
                report_failure = True
                if not self.nofail:
                    utils.die("reposync failed, retry limit reached, aborting")
                else:
                    self.logger.error("reposync failed, retry limit reached, skipping")

            self.update_permissions(repo_path)

        if report_failure:
            utils.die("overall reposync failed, at least one repo failed to synchronize")

    # ==================================================================================

    def sync(self, repo):

        """
        Conditionally sync a repo, based on type.

        :param repo: The repo to sync.
        """

        if repo.breed == RepoBreeds.RHN:
            self.rhn_sync(repo)
        elif repo.breed == RepoBreeds.YUM:
            self.yum_sync(repo)
        elif repo.breed == RepoBreeds.APT:
            self.apt_sync(repo)
        elif repo.breed == RepoBreeds.RSYNC:
            self.rsync_sync(repo)
        elif repo.breed == RepoBreeds.WGET:
            self.wget_sync(repo)
        else:
            utils.die("unable to sync repo (%s), unknown or unsupported repo type (%s)" % (repo.name, repo.breed.value))

    # ====================================================================================

    def librepo_getinfo(self, dirname):
        h = librepo.Handle()
        r = librepo.Result()
        h.setopt(librepo.LRO_REPOTYPE, librepo.LR_YUMREPO)
        h.setopt(librepo.LRO_URLS, [dirname])
        h.setopt(librepo.LRO_LOCAL, True)
        h.setopt(librepo.LRO_CHECKSUM, True)
        h.setopt(librepo.LRO_IGNOREMISSING, True)

        try:
            h.perform(r)
        except librepo.LibrepoException as e:
            utils.die("librepo error: " + dirname + " - " + e.args[1])

        rmd = r.getinfo(librepo.LRR_RPMMD_REPOMD)['records']
        return rmd

    # ====================================================================================

    def createrepo_walker(self, repo, dirname: str, fnames):
        """
        Used to run createrepo on a copied Yum mirror.

        :param repo: The repository object to run for.
        :param dirname: The directory to run in.
        :param fnames: Not known what this is for.
        """
        if os.path.exists(dirname) or repo.breed == RepoBreeds.RSYNC:
            utils.remove_yum_olddata(dirname)

            # add any repo metadata we can use
            mdoptions = []
            origin_path = os.path.join(dirname, ".origin")
            repodata_path = os.path.join(origin_path, "repodata")

            if os.path.isfile(os.path.join(repodata_path, "repomd.xml")):
                rd = self.librepo_getinfo(origin_path)

                if "group" in rd:
                    groupmdfile = rd['group']['location_href']
                    mdoptions.append("-g %s" % os.path.join(origin_path, groupmdfile))
                if "prestodelta" in rd:
                    # need createrepo >= 0.9.7 to add deltas
                    if utils.get_family() in ("redhat", "suse"):
                        cmd = "/usr/bin/rpmquery --queryformat=%{VERSION} createrepo"
                        createrepo_ver = utils.subprocess_get(cmd)
                        if not createrepo_ver[0:1].isdigit():
                            cmd = "/usr/bin/rpmquery --queryformat=%{VERSION} createrepo_c"
                            createrepo_ver = utils.subprocess_get(cmd)
                        if utils.compare_versions_gt(createrepo_ver, "0.9.7"):
                            mdoptions.append("--deltas")
                        else:
                            self.logger.error("this repo has presto metadata; you must upgrade createrepo to >= 0.9.7 "
                                              "first and then need to resync the repo through Cobbler.")

            blended = utils.blender(self.api, False, repo)
            flags = blended.get("createrepo_flags", "(ERROR: FLAGS)")
            try:
                cmd = "createrepo %s %s %s" % (" ".join(mdoptions), flags, pipes.quote(dirname))
                utils.subprocess_call(cmd)
            except:
                utils.log_exc()
                self.logger.error("createrepo failed.")
            del fnames[:]  # we're in the right place

    # ====================================================================================

    def wget_sync(self, repo):

        """
        Handle mirroring of directories using wget

        :param repo: The repo object to sync via wget.
        """

        repo_mirror = repo.mirror.strip()

        if repo.mirror_type != MirrorType.BASEURL:
            utils.die("mirrorlist and metalink mirror types is not supported for wget'd repositories")

        if repo.rpm_list != "" and repo.rpm_list != []:
            self.logger.warning("--rpm-list is not supported for wget'd repositories")

        # FIXME: don't hardcode
        dest_path = os.path.join(self.settings.webdir + "/repo_mirror", repo.name)

        # FIXME: wrapper for subprocess that logs to logger
        cmd = "wget -N -np -r -l inf -nd -P %s %s" % (pipes.quote(dest_path), pipes.quote(repo_mirror))
        rc = utils.subprocess_call(cmd)

        if rc != 0:
            utils.die("cobbler reposync failed")
        repo_walker(dest_path, self.createrepo_walker, repo)
        self.create_local_file(dest_path, repo)

    # ====================================================================================

    def rsync_sync(self, repo):

        """
        Handle copying of rsync:// and rsync-over-ssh repos.

        :param repo: The repo to sync via rsync.
        """

        if not repo.mirror_locally:
            utils.die("rsync:// urls must be mirrored locally, yum cannot access them directly")

        if repo.mirror_type != MirrorType.BASEURL:
            utils.die("mirrorlist and metalink mirror types is not supported for rsync'd repositories")

        if repo.rpm_list != "" and repo.rpm_list != []:
            self.logger.warning("--rpm-list is not supported for rsync'd repositories")

        # FIXME: don't hardcode
        dest_path = os.path.join(self.settings.webdir + "/repo_mirror", repo.name)

        spacer = ""
        if not repo.mirror.startswith("rsync://") and not repo.mirror.startswith("/"):
            spacer = "-e ssh"
        if not repo.mirror.strip().endswith("/"):
            repo.mirror = "%s/" % repo.mirror

        flags = ''
        for x in repo.rsyncopts:
            if repo.rsyncopts[x]:
                flags += " %s %s" % (x, repo.rsyncopts[x])
            else:
                flags += " %s" % x

        if flags == '':
            flags = self.settings.reposync_rsync_flags

        cmd = "rsync %s --delete-after %s --delete --exclude-from=/etc/cobbler/rsync.exclude %s %s" \
              % (flags, spacer, pipes.quote(repo.mirror), pipes.quote(dest_path))
        rc = utils.subprocess_call(cmd)

        if rc != 0:
            utils.die("cobbler reposync failed")

        # If ran in archive mode then repo should already contain all repodata and does not need createrepo run
        archive = False
        if '--archive' in flags:
            archive = True
        else:
            # split flags and skip all --{options} as we need to look for combined flags like -vaH
            fl = flags.split()
            for f in fl:
                if f.startswith('--'):
                    pass
                else:
                    if 'a' in f:
                        archive = True
                        break
        if not archive:
            repo_walker(dest_path, self.createrepo_walker, repo)

        self.create_local_file(dest_path, repo)

    # ====================================================================================

    def reposync_cmd(self) -> str:

        """
        Determine reposync command

        :return: The path to the reposync command. If dnf exists it is used instead of reposync.
        """

        cmd = None  # reposync command
        if os.path.exists("/usr/bin/dnf"):
            cmd = "/usr/bin/dnf reposync"
        elif os.path.exists("/usr/bin/reposync"):
            cmd = "/usr/bin/reposync"
        else:
            # Warn about not having yum-utils.  We don't want to require it in the package because Fedora 22+ has moved
            # to dnf.
            utils.die("no /usr/bin/reposync found, please install yum-utils")
        return cmd

    # ====================================================================================

    def rhn_sync(self, repo):

        """
        Handle mirroring of RHN repos.

        :param repo: The repo object to synchronize.
        """

        # reposync command
        cmd = self.reposync_cmd()

        # flag indicating not to pull the whole repo
        has_rpm_list = False

        # detect cases that require special handling
        if repo.rpm_list != "" and repo.rpm_list != []:
            has_rpm_list = True

        # Create yum config file for use by reposync
        # FIXME: don't hardcode
        dest_path = os.path.join(self.settings.webdir + "/repo_mirror", repo.name)
        temp_path = os.path.join(dest_path, ".origin")

        if not os.path.isdir(temp_path):
            # FIXME: there's a chance this might break the RHN D/L case
            os.makedirs(temp_path)

        # how we invoke reposync depends on whether this is RHN content or not.

        # This is the somewhat more-complex RHN case.
        # NOTE: this requires that you have entitlements for the server and you give the mirror as rhn://$channelname
        if not repo.mirror_locally:
            utils.die("rhn:// repos do not work with --mirror-locally=False")

        if has_rpm_list:
            self.logger.warning("warning: --rpm-list is not supported for RHN content")
        rest = repo.mirror[6:]  # everything after rhn://
        cmd = "%s %s --repo=%s -p %s" % (cmd,
                                         self.rflags,
                                         pipes.quote(rest),
                                         pipes.quote(self.settings.webdir + "/repo_mirror"))
        if repo.name != rest:
            args = {"name": repo.name, "rest": rest}
            utils.die("ERROR: repository %(name)s needs to be renamed %(rest)s as the name of the cobbler repository "
                      "must match the name of the RHN channel" % args)

        arch = repo.arch.value

        if arch == "i386":
            # Counter-intuitive, but we want the newish kernels too
            arch = "i686"

        if arch != "none":
            cmd = "%s -a %s" % (cmd, arch)

        # Now regardless of whether we're doing yumdownloader or reposync or whether the repo was http://, ftp://, or
        # rhn://, execute all queued commands here. Any failure at any point stops the operation.

        if repo.mirror_locally:
            utils.subprocess_call(cmd)

        # Some more special case handling for RHN. Create the config file now, because the directory didn't exist
        # earlier.

        self.create_local_file(temp_path, repo, output=False)

        # Now run createrepo to rebuild the index

        if repo.mirror_locally:
            repo_walker(dest_path, self.createrepo_walker, repo)

        # Create the config file the hosts will use to access the repository.

        self.create_local_file(dest_path, repo)

    # ====================================================================================

    def gen_urlgrab_ssl_opts(self, yumopts) -> Union[str, bool]:
        """
        This function translates yum repository options into the appropriate options for python-requests

        :param yumopts: The options to convert.
        :return: A tuple with the cert and a boolean if it should be verified or not.
        """
        # use SSL options if specified in yum opts
        cert = None
        sslcacert = None
        verify = False
        if 'sslcacert' in yumopts:
            sslcacert = yumopts['sslcacert']
        if 'sslclientkey' and 'sslclientcert' in yumopts:
            cert = (sslcacert, yumopts['sslclientcert'], yumopts['sslclientkey'])
        # Note that the default of requests is to verify the peer and host but the default here is NOT to verify them
        # unless sslverify is explicitly set to 1 in yumopts.
        if 'sslverify' in yumopts:
            if yumopts['sslverify'] == 1:
                verify = True
            else:
                verify = False

        return (cert, verify)

    # ====================================================================================

    def yum_sync(self, repo):

        """
        Handle copying of http:// and ftp:// yum repos.

        :param repo: The yum reporitory to sync.
        """

        # create the config file the hosts will use to access the repository.
        repo_mirror = repo.mirror.strip()
        dest_path = os.path.join(self.settings.webdir + "/repo_mirror", repo.name.strip())
        self.create_local_file(dest_path, repo)

        if not repo.mirror_locally:
            return

        # command to run
        cmd = self.reposync_cmd()
        # flag indicating not to pull the whole repo
        has_rpm_list = False

        # detect cases that require special handling
        if repo.rpm_list != "" and repo.rpm_list != []:
            has_rpm_list = True

        # create yum config file for use by reposync
        temp_path = os.path.join(dest_path, ".origin")

        if not os.path.isdir(temp_path):
            # FIXME: there's a chance this might break the RHN D/L case
            os.makedirs(temp_path)

        temp_file = self.create_local_file(temp_path, repo, output=False)

        arch = repo.arch.value
        if arch == "i386":
            # Counter-intuitive, but we want the newish kernels too
            arch = "i686"

        if not has_rpm_list:
            # If we have not requested only certain RPMs, use reposync
            cmd = "%s %s --config=%s --repoid=%s -p %s" \
                  % (cmd, self.rflags, temp_file, pipes.quote(repo.name),
                     pipes.quote(self.settings.webdir + "/repo_mirror"))
            if arch != "none":
                cmd = "%s -a %s" % (cmd, arch)

        else:
            # Create the output directory if it doesn't exist
            if not os.path.exists(dest_path):
                os.makedirs(dest_path)

            use_source = ""
            if arch == "src":
                use_source = "--source"

            # Older yumdownloader sometimes explodes on --resolvedeps if this happens to you, upgrade yum & yum-utils
            extra_flags = self.settings.yumdownloader_flags
            cmd = "/usr/bin/dnf download"
            cmd = "%s %s %s --disablerepo=* --enablerepo=%s -c %s --destdir=%s %s" \
                  % (cmd, extra_flags, use_source, pipes.quote(repo.name), temp_file, pipes.quote(dest_path),
                     " ".join(repo.rpm_list))

        # Now regardless of whether we're doing yumdownloader or reposync or whether the repo was http://, ftp://, or
        # rhn://, execute all queued commands here.  Any failure at any point stops the operation.

        rc = utils.subprocess_call(cmd)
        if rc != 0:
            utils.die("cobbler reposync failed")

        # download any metadata we can use
        proxy = None
        if repo.proxy != '<<None>>' and repo.proxy != '':
            proxy = repo.proxy
        (cert, verify) = self.gen_urlgrab_ssl_opts(repo.yumopts)

        repodata_path = os.path.join(dest_path, "repodata")
        repomd_path = os.path.join(repodata_path, "repomd.xml")

        if os.path.exists(repodata_path) and not os.path.isfile(repomd_path):
            shutil.rmtree(repodata_path, ignore_errors=False, onerror=None)

        h = librepo.Handle()
        r = librepo.Result()
        h.setopt(librepo.LRO_REPOTYPE, librepo.LR_YUMREPO)
        h.setopt(librepo.LRO_CHECKSUM, True)

        if os.path.isfile(repomd_path):
            h.setopt(librepo.LRO_LOCAL, True)
            h.setopt(librepo.LRO_URLS, [temp_path])
            h.setopt(librepo.LRO_IGNOREMISSING, True)

            try:
                h.perform(r)
            except librepo.LibrepoException as e:
                utils.die("librepo error: " + temp_path + " - " + e.args[1])

            h.setopt(librepo.LRO_LOCAL, False)
            h.setopt(librepo.LRO_URLS, [])
            h.setopt(librepo.LRO_IGNOREMISSING, False)
            h.setopt(librepo.LRO_UPDATE, True)

        h.setopt(librepo.LRO_DESTDIR, temp_path)

        if repo.mirror_type == MirrorType.METALINK:
            h.setopt(librepo.LRO_METALINKURL, repo_mirror)
        elif repo.mirror_type == MirrorType.MIRRORLIST:
            h.setopt(librepo.LRO_MIRRORLISTURL, repo_mirror)
        elif repo.mirror_type == MirrorType.BASEURL:
            h.setopt(librepo.LRO_URLS, [repo_mirror])

        if verify:
            h.setopt(librepo.LRO_SSLVERIFYPEER, True)
            h.setopt(librepo.LRO_SSLVERIFYHOST, True)

        if cert:
            sslcacert, sslclientcert, sslclientkey = cert
            h.setopt(librepo.LRO_SSLCACERT, sslcacert)
            h.setopt(librepo.LRO_SSLCLIENTCERT, sslclientcert)
            h.setopt(librepo.LRO_SSLCLIENTKEY, sslclientkey)

        if proxy:
            h.setopt(librepo.LRO_PROXY, proxy)
            h.setopt(librepo.LRO_PROXYTYPE, librepo.PROXY_HTTP)

        try:
            h.perform(r)
        except librepo.LibrepoException as e:
            utils.die("librepo error: " + temp_path + " - " + e.args[1])

        # now run createrepo to rebuild the index
        if repo.mirror_locally:
            repo_walker(dest_path, self.createrepo_walker, repo)

    # ====================================================================================

    def apt_sync(self, repo):
        """
        Handle copying of http:// and ftp:// debian repos.

        :param repo: The apt repository to sync.
        """

        # Warn about not having mirror program.
        mirror_program = "/usr/bin/debmirror"
        if not os.path.exists(mirror_program):
            utils.die("no %s found, please install it" % (mirror_program))

        # command to run
        cmd = ""

        # detect cases that require special handling
        if repo.rpm_list != "" and repo.rpm_list != []:
            utils.die("has_rpm_list not yet supported on apt repos")

        if repo.arch == RepoArchs.NONE:
            utils.die("Architecture is required for apt repositories")

        if repo.mirror_type != MirrorType.BASEURL:
            utils.die("mirrorlist and metalink mirror types is not supported for apt repositories")

        # built destination path for the repo
        dest_path = os.path.join("/var/www/cobbler/repo_mirror", repo.name)

        if repo.mirror_locally:
            # NOTE: Dropping @@suite@@ replace as it is also dropped from "from manage_import_debian"_ubuntu.py due that
            # repo has no os_version attribute. If it is added again it will break the Web UI!
            # mirror = repo.mirror.replace("@@suite@@",repo.os_version)
            mirror = repo.mirror

            idx = mirror.find("://")
            method = mirror[:idx]
            mirror = mirror[idx + 3:]

            idx = mirror.find("/")
            host = mirror[:idx]
            mirror = mirror[idx:]

            dists = ",".join(repo.apt_dists)
            components = ",".join(repo.apt_components)

            mirror_data = "--method=%s --host=%s --root=%s --dist=%s --section=%s" \
                          % (pipes.quote(method), pipes.quote(host), pipes.quote(mirror), pipes.quote(dists),
                             pipes.quote(components))

            rflags = "--nocleanup"
            for x in repo.yumopts:
                if repo.yumopts[x]:
                    rflags += " %s=%s" % (x, repo.yumopts[x])
                else:
                    rflags += " %s" % x
            cmd = "%s %s %s %s" % (mirror_program, rflags, mirror_data, dest_path)
            cmd = "%s %s %s %s" % (mirror_program, rflags, mirror_data, pipes.quote(dest_path))
            if repo.arch == RepoArchs.SRC:
                cmd = "%s --source" % cmd
            else:
                arch = repo.arch.value
                if arch == "x86_64":
                    arch = "amd64"  # FIX potential arch errors
                cmd = "%s --nosource -a %s" % (cmd, arch)

            # Set's an environment variable for subprocess, otherwise debmirror will fail as it needs this variable to
            # exist.
            # FIXME: might this break anything? So far it doesn't
            os.putenv("HOME", "/var/lib/cobbler")

            rc = utils.subprocess_call(cmd)
            if rc != 0:
                utils.die("cobbler reposync failed")

    def create_local_file(self, dest_path: str, repo, output: bool = True):
        """
        Creates Yum config files for use by reposync

        Two uses:
        (A) output=True, Create local files that can be used with yum on provisioned clients to make use of this mirror.
        (B) output=False, Create a temporary file for yum to feed into yum for mirroring

        :param dest_path: The destination path to create the file at.
        :param repo: The repository object to create a file for.
        :param output: See described above.
        :return: The name of the file which was written.
        """

        # The output case will generate repo configuration files which are usable for the installed systems. They need
        # to be made compatible with --server-override which means they are actually templates, which need to be
        # rendered by a Cobbler-sync on per profile/system basis.

        if output:
            fname = os.path.join(dest_path, "config.repo")
        else:
            fname = os.path.join(dest_path, "%s.repo" % repo.name)
        self.logger.debug("creating: %s" % fname)
        if not os.path.exists(dest_path):
            utils.mkdir(dest_path)
        config_file = open(fname, "w+")
        if not output:
            config_file.write("[main]\nreposdir=/dev/null\n")
        config_file.write("[%s]\n" % repo.name)
        config_file.write("name=%s\n" % repo.name)

        optenabled = False
        optgpgcheck = False
        if output:
            if repo.mirror_locally:
                line = "baseurl=http://${http_server}/cobbler/repo_mirror/%s\n" % (repo.name)
            else:
                mstr = repo.mirror
                if mstr.startswith("/"):
                    mstr = "file://%s" % mstr
                line = "%s=%s\n" % (repo.mirror_type.value, mstr)

            config_file.write(line)
            # User may have options specific to certain yum plugins add them to the file
            for x in repo.yumopts:
                if x == "enabled":
                    optenabled = True
                if x == "gpgcheck":
                    optgpgcheck = True
        else:
            mstr = repo.mirror
            if mstr.startswith("/"):
                mstr = "file://%s" % mstr
            line = repo.mirror_type.value + "=%s\n" % mstr
            if self.settings.http_port not in (80, '80'):
                http_server = "%s:%s" % (self.settings.server, self.settings.http_port)
            else:
                http_server = self.settings.server
            line = line.replace("@@server@@", http_server)
            config_file.write(line)

            config_proxy = None
            if repo.proxy == '<<inherit>>':
                config_proxy = self.settings.proxy_url_ext
            elif repo.proxy != '' and repo.proxy != '<<None>>':
                config_proxy = repo.proxy

            if config_proxy is not None:
                config_file.write("proxy=%s\n" % config_proxy)
            if 'exclude' in list(repo.yumopts.keys()):
                self.logger.debug("excluding: %s" % repo.yumopts['exclude'])
                config_file.write("exclude=%s\n" % repo.yumopts['exclude'])

        if not optenabled:
            config_file.write("enabled=1\n")
        config_file.write("priority=%s\n" % repo.priority)
        # FIXME: potentially might want a way to turn this on/off on a per-repo basis
        if not optgpgcheck:
            config_file.write("gpgcheck=0\n")
        # user may have options specific to certain yum plugins
        # add them to the file
        for x in repo.yumopts:
            if not (output and repo.mirror_locally and x.startswith("ssl")):
                config_file.write("%s=%s\n" % (x, repo.yumopts[x]))
        config_file.close()
        return fname

    # ==================================================================================

    def update_permissions(self, repo_path):
        """
        Verifies that permissions and contexts after an rsync are as expected.
        Sending proper rsync flags should prevent the need for this, though this is largely a safeguard.

        :param repo_path: The path to update the permissions of.
        """
        # all_path = os.path.join(repo_path, "*")
        owner = "root:apache"

        (dist, _) = os_release()
        if dist == "suse":
            owner = "root:www"
        elif dist in ("debian", "ubuntu"):
            owner = "root:www-data"

        cmd1 = "chown -R " + owner + " %s" % repo_path

        utils.subprocess_call(cmd1)

        cmd2 = "chmod -R 755 %s" % repo_path
        utils.subprocess_call(cmd2)



# coding=utf-8
"""
This is some of the code behind 'cobbler sync'.

Copyright 2014, Mittwald CM Service GmbH & Co. KG
Martin Helmich <m.helmich@mittwald.de>
Daniel Krämer <d.kraemer@mittwald.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""
import os
import subprocess

from cobbler.manager import ManagerModule

MANAGER = None


def register() -> str:
    """
    The mandatory Cobbler module registration hook.
    """
    return "manage"


class _NDjbDnsManager(ManagerModule):

    @staticmethod
    def what() -> str:
        """
        Static method to identify the manager.

        :return: Always "ndjbdns".
        """
        return "ndjbdns"

    def __init__(self, collection_mgr):
        super().__init__(collection_mgr)

    def write_configs(self):
        """
        This writes the new dns configuration file to the disc.
        """
        template_file = '/etc/cobbler/ndjbdns.template'
        data_file = '/etc/ndjbdns/data'
        data_dir = os.path.dirname(data_file)

        a_records = {}

        with open(template_file, 'r') as f:
            template_content = f.read()

        for system in self.systems:
            for (name, interface) in list(system.interfaces.items()):
                host = interface['dns_name']
                ip = interface['ip_address']

                if host:
                    if host in a_records:
                        raise Exception('Duplicate DNS name: %s' % host)
                    a_records[host] = ip

        template_vars = {'forward': []}
        for host, ip in list(a_records.items()):
            template_vars['forward'].append((host, ip))

        self.templar.render(template_content, template_vars, data_file)

        p = subprocess.Popen(['/usr/bin/tinydns-data'], cwd=data_dir)
        p.communicate()

        if p.returncode != 0:
            raise Exception('Could not regenerate tinydns data file.')


def get_manager(collection_mgr):
    """
    Creates a manager object to manage an isc dhcp server.

    :param collection_mgr: The collection manager which holds all information in the current Cobbler instance.
    :return: The object to manage the server with.
    """
    # Singleton used, therefore ignoring 'global'
    global MANAGER  # pylint: disable=global-statement

    if not MANAGER:
        MANAGER = _NDjbDnsManager(collection_mgr)
    return MANAGER



import logging
import os
import sys
import xmlrpc.client as xmlrpcclient
from pathlib import Path

import pytest

from cobbler.utils import local_get_cobbler_api_url, get_shared_secret

# "import xmlrpc.client" does currently not work. No explanation found anywhere.


@pytest.fixture(scope="session")
def remote(cobbler_xmlrpc_base) -> xmlrpcclient.ServerProxy:
    """

    :param cobbler_xmlrpc_base:
    :return:
    """
    return cobbler_xmlrpc_base[0]


@pytest.fixture(scope="session")
def token(cobbler_xmlrpc_base) -> str:
    """

    :param cobbler_xmlrpc_base:
    :return:
    """
    return cobbler_xmlrpc_base[1]


@pytest.fixture(scope="session")
def cobbler_xmlrpc_base():
    """
    Initialises the api object and makes it available to the test.
    """
    # create logger
    logging.basicConfig(stream=sys.stderr)
    logger = logging.getLogger("xobbler_xmlrpc_base")
    logger.setLevel(logging.DEBUG)

    # create XML-RPC client and connect to server
    api_url = local_get_cobbler_api_url()
    remote = xmlrpcclient.Server(api_url, allow_none=True)
    shared_secret = get_shared_secret()
    token = remote.login("", shared_secret)
    if not token:
        sys.exit(1)
    yield remote, token


@pytest.fixture(scope="class")
def testsnippet() -> str:
    return "# This is a small simple testsnippet!"


@pytest.fixture()
def snippet_add(remote, token):
    def _snippet_add(name: str, data):
        remote.write_autoinstall_snippet(name, data, token)
    return _snippet_add


@pytest.fixture()
def snippet_remove(remote, token):
    def _snippet_remove(name: str):
        remote.remove_autoinstall_snippet(name, token)
    return _snippet_remove


@pytest.fixture()
def create_distro(remote, token):
    def _create_distro(name: str, arch: str, breed: str, path_kernel: str, path_initrd: str):
        distro = remote.new_distro(token)
        remote.modify_distro(distro, "name", name, token)
        remote.modify_distro(distro, "arch", arch, token)
        remote.modify_distro(distro, "breed", breed, token)
        remote.modify_distro(distro, "kernel", path_kernel, token)
        remote.modify_distro(distro, "initrd", path_initrd, token)
        remote.save_distro(distro, token)
        return distro
    return _create_distro


@pytest.fixture()
def remove_distro(remote, token):
    def _remove_distro(name: str):
        remote.remove_distro(name, token)
    return _remove_distro


@pytest.fixture()
def create_profile(remote, token):
    def _create_profile(name, distro, kernel_options):
        profile = remote.new_profile(token)
        remote.modify_profile(profile, "name", name, token)
        remote.modify_profile(profile, "distro", distro, token)
        remote.modify_profile(profile, "kernel_options", kernel_options, token)
        remote.save_profile(profile, token)
        return profile
    return _create_profile


@pytest.fixture()
def remove_profile(remote, token):
    def _remove_profile(name):
        remote.remove_profile(name, token)
    return _remove_profile


@pytest.fixture()
def create_system(remote, token):
    def _create_system(name, profile):
        system = remote.new_system(token)
        remote.modify_system(system, "name", name, token)
        remote.modify_system(system, "profile", profile, token)
        remote.save_system(system, token)
        return system
    return _create_system


@pytest.fixture()
def remove_system(remote, token):
    def _remove_system(name):
        remote.remove_system(name, token)
    return _remove_system


@pytest.fixture()
def create_file(remote, token):
    def _create_file(name, is_directory, action, group, mode, owner, path, template):
        file_id = remote.new_file(token)

        remote.modify_file(file_id, "name", name, token)
        remote.modify_file(file_id, "is_directory", is_directory, token)
        remote.modify_file(file_id, "action", action, token)
        remote.modify_file(file_id, "group", group, token)
        remote.modify_file(file_id, "mode", mode, token)
        remote.modify_file(file_id, "owner", owner, token)
        remote.modify_file(file_id, "path", path, token)
        remote.modify_file(file_id, "template", template, token)

        remote.save_file(file_id, token)
        return file_id
    return _create_file


@pytest.fixture()
def remove_file(remote, token):
    def _remove_file(name):
        remote.remove_file(name, token)
    return _remove_file


@pytest.fixture()
def create_mgmt_class(remote, token):
    def _create_mgmt_class(name):
        mgmtclass = remote.new_mgmtclass(token)
        remote.modify_mgmtclass(mgmtclass, "name", name, token)
        remote.save_mgmtclass(mgmtclass, token)
        return mgmtclass
    return _create_mgmt_class


@pytest.fixture()
def remove_mgmt_class(remote, token):
    def _remove_mgmt_class(name):
        remote.remove_mgmtclass(name, token)
    return _remove_mgmt_class


@pytest.fixture()
def create_autoinstall_template(remote, token):
    def _create_autoinstall_template(filename, content):
        remote.write_autoinstall_template(filename, content, token)
    return _create_autoinstall_template


@pytest.fixture()
def remove_autoinstall_template(remote, token):
    def _remove_autoinstall_template(name):
        remote.remove_autoinstall_template(name, token)
    return _remove_autoinstall_template


@pytest.fixture
def create_repo(remote, token):
    def _create_repo(name, mirror, mirror_locally):
        repo = remote.new_repo(token)
        remote.modify_repo(repo, "name", name, token)
        remote.modify_repo(repo, "mirror", mirror, token)
        remote.modify_repo(repo, "mirror_locally", mirror_locally, token)
        remote.save_repo(repo, token)
        return repo
    return _create_repo


@pytest.fixture
def remove_repo(remote, token):
    def _remove_repo(name):
        remote.remove_repo(name, token)
    return _remove_repo


@pytest.fixture()
def create_menu(remote, token):
    def _create_menu(name, display_name):
        menu_id = remote.new_menu(token)

        remote.modify_menu(menu_id, "name", name, token)
        remote.modify_menu(menu_id, "display_name", display_name, token)

        remote.save_menu(menu_id, token)
        return menu_id
    return _create_menu


@pytest.fixture()
def remove_menu(remote, token):
    def _remove_menu(name):
        remote.remove_menu(name, token)
    return _remove_menu


@pytest.fixture(scope="function")
def create_testprofile(remote, token):
    """
    Create a profile with the name "testprofile0"
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    profile = remote.new_profile(token)
    remote.modify_profile(profile, "name", "testprofile0", token)
    remote.modify_profile(profile, "distro", "testdistro0", token)
    remote.modify_profile(profile, "kernel_options", "a=1 b=2 c=3 c=4 c=5 d e", token)
    remote.modify_profile(profile, "menu", "testmenu0", token)
    remote.save_profile(profile, token)


@pytest.fixture(scope="function")
def remove_testprofile(remote, token):
    """
    Removes the profile with the name "testprofile0".
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    yield
    remote.remove_profile("testprofile0", token)


@pytest.fixture(scope="function")
def remove_testdistro(remote, token):
    """
    Removes the distro "testdistro0" from the running cobbler after the test.
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    yield
    remote.remove_distro("testdistro0", token, False)


@pytest.fixture(scope="function")
def create_testdistro(remote, token, fk_kernel, fk_initrd, create_kernel_initrd):
    """
    Creates a distro "testdistro0" with the architecture "x86_64", breed "suse" and the fixtures which are setting the
    fake kernel and initrd.
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    :param fk_kernel: See the corresponding fixture.
    :param fk_initrd: See the corresponding fixture.
    """
    folder = create_kernel_initrd(fk_kernel, fk_initrd)
    distro = remote.new_distro(token)
    remote.modify_distro(distro, "name", "testdistro0", token)
    remote.modify_distro(distro, "arch", "x86_64", token)
    remote.modify_distro(distro, "breed", "suse", token)
    remote.modify_distro(distro, "kernel", os.path.join(folder, fk_kernel), token)
    remote.modify_distro(distro, "initrd", os.path.join(folder, fk_initrd), token)
    remote.save_distro(distro, token)


@pytest.fixture(scope="function")
def create_testsystem(remote, token):
    """
    Add a system with the name "testsystem0", the system is assigend to the profile "testprofile0".
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    system = remote.new_system(token)
    remote.modify_system(system, "name", "testsystem0", token)
    remote.modify_system(system, "profile", "testprofile0", token)
    remote.save_system(system, token)


@pytest.fixture()
def remove_testsystem(remote, token):
    """
    Remove a system "testsystem0".
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    yield
    remote.remove_system("testsystem0", token, False)


@pytest.fixture(scope="function")
def create_testrepo(remote, token):
    """
    Create a testrepository with the name "testrepo0"
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    repo = remote.new_repo(token)
    remote.modify_repo(repo, "name", "testrepo0", token)
    remote.modify_repo(repo, "arch", "x86_64", token)
    remote.modify_repo(repo, "mirror", "http://something", token)
    remote.save_repo(repo, token)


@pytest.fixture(scope="function")
def remove_testrepo(remote, token):
    """
    Remove a repo "testrepo0".
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    yield
    remote.remove_repo("testrepo0", token, False)


@pytest.fixture(scope="function")
def create_testimage(remote, token):
    """
    Create a testrepository with the name "testimage0"
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    image = remote.new_image(token)
    remote.modify_image(image, "name", "testimage0", token)
    remote.save_image(image, token)


@pytest.fixture(scope="function")
def remove_testimage(remote, token):
    """
    Remove the image "testimage0".
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    yield
    remote.remove_image("testimage0", token, False)


@pytest.fixture(scope="function")
def create_testpackage(remote, token):
    """
    Create a testpackage with the name "testpackage0"
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    package = remote.new_package(token)
    remote.modify_package(package, "name", "testpackage0", token)
    remote.save_package(package, token)


@pytest.fixture(scope="function")
def remove_testpackage(remote, token):
    """
    Remove a package "testpackage0".
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """

    yield
    remote.remove_package("testpackage0", token, False)


@pytest.fixture(scope="function")
def create_testfile_item(remote, token):
    """
    Create a testfile with the name "testfile0"
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """

    mfile = remote.new_file(token)
    remote.modify_file(mfile, "name", "testfile0", token)
    remote.modify_file(mfile, "path", "/dev/shm/", token)
    remote.modify_file(mfile, "group", "root", token)
    remote.modify_file(mfile, "owner", "root", token)
    remote.modify_file(mfile, "mode", "0600", token)
    remote.modify_file(mfile, "is_dir", "True", token)
    remote.save_file(mfile, token)


@pytest.fixture(scope="function")
def remove_testfile(remote, token):
    """
    Remove a file "testfile0".
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    yield
    remote.remove_file("testfile0", token, False)


@pytest.fixture(scope="function")
def create_mgmtclass(remote, token):
    """
    Create a mgmtclass with the name "mgmtclass0"
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """

    mgmtclass0 = remote.new_mgmtclass(token)
    remote.modify_mgmtclass(mgmtclass0, "name", "mgmtclass0", token)
    remote.save_mgmtclass(mgmtclass0, token)


@pytest.fixture(scope="function")
def remove_mgmtclass(remote, token):
    """
    Remove a mgmtclass "mgmtclass0".
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    yield
    remote.remove_mgmtclass("mgmtclass0", token, False)


@pytest.fixture(scope="function")
def create_testmenu(remote, token):
    """
    Create a menu with the name "testmenu0"
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """

    menu = remote.new_menu(token)
    remote.modify_menu(menu, "name", "testmenu0", token)
    remote.save_menu(menu, token)


@pytest.fixture(scope="function")
def remove_testmenu(remote, token):
    """
    Remove a menu "testmenu0".
    :param remote: See the corresponding fixture.
    :param token: See the corresponding fixture.
    """
    yield
    remote.remove_menu("testmenu0", token, False)


@pytest.fixture(scope="function")
def template_files(redhat_autoinstall, suse_autoyast, ubuntu_preseed):
    """
    Create the template files and remove them afterwards.

    :return:
    """
    folder = "/var/lib/cobbler/templates"
    Path(os.path.join(folder, redhat_autoinstall)).touch()
    Path(os.path.join(folder, suse_autoyast)).touch()
    Path(os.path.join(folder, ubuntu_preseed)).touch()

    yield

    os.remove(os.path.join(folder, redhat_autoinstall))
    os.remove(os.path.join(folder, suse_autoyast))
    os.remove(os.path.join(folder, ubuntu_preseed))



import pytest

from cobbler import grub
from tests.conftest import does_not_raise


@pytest.mark.parametrize("input_file_location,expected_output,expected_exception", [
    (None, None, pytest.raises(TypeError)),
    ("ftp://testuri", None, does_not_raise()),
    ("http://testuri", None, pytest.raises(ValueError)),
    ("tftp://testuri", None, pytest.raises(ValueError)),
    ("wss://testuri", None, does_not_raise()),
    ("http://10.0.0.1", "(http,10.0.0.1)/", does_not_raise()),
    ("tftp://10.0.0.1", "(tftp,10.0.0.1)/", does_not_raise()),
    ("tftp://10.0.0.1/testpath/testpath/", "(tftp,10.0.0.1)/testpath/testpath/", does_not_raise()),
])
def test_parse_grub_remote_file(input_file_location, expected_output, expected_exception):
    # Arrange & Act
    with expected_exception:
        result = grub.parse_grub_remote_file(input_file_location)

        # Assert
        assert result == expected_output



"""
Power management library.  Encapsulate the logic to run power management
commands so that the Cobbler user does not have to remember different power
management tools syntaxes.  This makes rebooting a system for OS installation
much easier.

Copyright 2008-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""
import json
import glob
import logging
import os
from pathlib import Path
import stat
import re
import time
from typing import Optional

from cobbler.cexceptions import CX
from cobbler import utils

# Try the power command 3 times before giving up. Some power switches are flaky.
POWER_RETRIES = 3


def get_power_types() -> list:
    """
    Get possible power management types.

    :returns: Possible power management types
    """

    power_types = []
    fence_files = glob.glob("/usr/sbin/fence_*") + glob.glob("/sbin/fence_*")
    for x in fence_files:
        fence_name = os.path.basename(x).replace("fence_", "")
        if fence_name not in power_types:
            power_types.append(fence_name)
    power_types.sort()
    return power_types


def validate_power_type(power_type: str):
    """
    Check if a power management type is valid.

    :param power_type: Power management type.
    :raise CX: if power management type is invalid
    """

    power_types = get_power_types()
    if not power_types:
        raise CX("you need to have fence-agents installed")
    if power_type not in power_types:
        raise CX("power management type must be one of: %s" % ",".join(power_types))


def get_power_command(power_type: str) -> Optional[str]:
    """
    Get power management command path

    :param power_type: power management type
    :returns: power management command path
    """

    if power_type:
        # try /sbin, then /usr/sbin
        power_path1 = "/sbin/fence_%s" % power_type
        power_path2 = "/usr/sbin/fence_%s" % power_type
        for power_path in (power_path1, power_path2):
            if os.path.isfile(power_path) and os.access(power_path, os.X_OK):
                return power_path
    return None


class PowerManager:
    """
    Handles power management in systems
    """

    def __init__(self, api, collection_mgr):
        """
        Constructor

        :param api: Cobbler API
        :type api: CobblerAPI
        :param collection_mgr: collection manager
        :type collection_mgr: CollectionManager
        """

        self.collection_mgr = collection_mgr
        self.settings = collection_mgr.settings()
        self.api = api
        self.logger = logging.getLogger()

    def _check_power_conf(self, system, user, password):
        """
        Prints a warning for invalid power configurations.

        :param user: The username for the power command of the system. This overrules the one specified in the system.
        :param password: The password for the power command of the system. This overrules the one specified in the
                         system.
        :param system: Cobbler system
        :type system: System
        """

        if (system.power_pass or password) and system.power_identity_file:
            self.logger.warning("Both password and identity-file are specified")
        if system.power_identity_file:
            ident_path = Path(system.power_identity_file)
            if not ident_path.exists():
                self.logger.warning("identity-file " + system.power_identity_file + " does not exist")
            else:
                ident_stat = stat.S_IMODE(ident_path.stat().st_mode)
                if (ident_stat & stat.S_IRWXO) or (ident_stat & stat.S_IRWXG):
                    self.logger.warning("identity-file %s must not be read/write/exec by group or others",
                                        system.power_identity_file)
        if not system.power_address:
            self.logger.warning("power-address is missing")
        if not (system.power_user or user):
            self.logger.warning("power-user is missing")
        if not (system.power_pass or password) and not system.power_identity_file:
            self.logger.warning("neither power-identity-file nor power-password specified")

    def _get_power_input(self, system, power_operation: str, user: str, password: str) -> str:
        """
        Creates an option string for the fence agent from the system data. This is an internal method.

        :param system: Cobbler system
        :type system: System
        :param power_operation: power operation. Valid values: on, off, status. Rebooting is implemented as a set of 2
                                operations (off and on) in a higher level method.
        :param user: user to override system.power_user
        :param password: password to override system.power_pass
        :return: The option string for the fencer agent.
        """

        self._check_power_conf(system, user, password)
        power_input = ""
        if power_operation is None or power_operation not in ['on', 'off', 'status']:
            raise CX("invalid power operation")
        power_input += "action=" + power_operation + "\n"
        if system.power_address:
            power_input += "ip=" + system.power_address + "\n"
        if system.power_user:
            power_input += "username=" + system.power_user + "\n"
        if system.power_id:
            power_input += "plug=" + system.power_id + "\n"
        if system.power_pass:
            power_input += "password=" + system.power_pass + "\n"
        if system.power_identity_file:
            power_input += "identity-file=" + system.power_identity_file + "\n"
        if system.power_options:
            power_input += system.power_options + "\n"
        return power_input

    def _power(self, system, power_operation: str, user: Optional[str] = None,
               password: Optional[str] = None) -> Optional[bool]:
        """
        Performs a power operation on a system.
        Internal method

        :param system: Cobbler system
        :type system: System
        :param power_operation: power operation. Valid values: on, off, status. Rebooting is implemented as a set of 2
                                operations (off and on) in a higher level method.
        :param user: power management user. If user and password are not supplied, environment variables
                     COBBLER_POWER_USER and COBBLER_POWER_PASS will be used.
        :param password: power management password
        :return: bool/None if power operation is 'status', return if system is on; otherwise, return None
        :raise CX: if there are errors
        """

        power_command = get_power_command(system.power_type)
        if not power_command:
            utils.die("no power type set for system")

        power_info = {"type": system.power_type, "address": system.power_address, "user": system.power_user,
                      "id": system.power_id, "options": system.power_options,
                      "identity_file": system.power_identity_file}

        self.logger.info("cobbler power configuration is: %s", json.dumps(power_info))

        # if no username/password data, check the environment
        if not system.power_user and not user:
            user = os.environ.get("COBBLER_POWER_USER", "")
        if not system.power_pass and not password:
            password = os.environ.get("COBBLER_POWER_PASS", "")

        power_input = self._get_power_input(system, power_operation, user, password)

        self.logger.info("power command: %s", power_command)
        self.logger.info("power command input: %s", power_input)

        rc = -1

        for x in range(0, POWER_RETRIES):
            output, rc = utils.subprocess_sp(power_command, shell=False, input=power_input)
            # Allowed return codes: 0, 1, 2
            # Source: https://github.com/ClusterLabs/fence-agents/blob/master/doc/FenceAgentAPI.md#agent-operations-and-return-values
            if power_operation in ("on", "off", "reboot"):
                if rc == 0:
                    return None
            elif power_operation == "status":
                if rc in (0, 2):
                    match = re.match(r'^(Status:|.+power\s=)\s(on|off)$', output, re.IGNORECASE | re.MULTILINE)
                    if match:
                        power_status = match.groups()[1]
                        if power_status.lower() == 'on':
                            return True
                        else:
                            return False
                    error_msg = "command succeeded (rc=%s), but output ('%s') was not understood" % (rc, output)
                    utils.die(error_msg)
                    raise CX(error_msg)
            time.sleep(2)

        if not rc == 0:
            error_msg = "command failed (rc=%s), please validate the physical setup and cobbler config" % rc
            utils.die(error_msg)
            raise CX(error_msg)

    def power_on(self, system, user: Optional[str] = None, password: Optional[str] = None):
        """
        Powers up a system that has power management configured.

        :param system: Cobbler system
        :type system: System
        :param user: power management user
        :param password: power management password
        """

        self._power(system, "on", user, password)

    def power_off(self, system, user: Optional[str] = None, password: Optional[str] = None):
        """
        Powers down a system that has power management configured.

        :param system: Cobbler system
        :type system: System
        :param user: power management user
        :param password: power management password
        """

        self._power(system, "off", user, password)

    def reboot(self, system, user: Optional[str] = None, password: Optional[str] = None):
        """
        Reboot a system that has power management configured.

        :param system: Cobbler system
        :type system: System
        :param user: power management user
        :param password: power management password
        """

        self.power_off(system, user, password)
        time.sleep(5)
        self.power_on(system, user, password)

    def get_power_status(self, system, user: Optional[str] = None, password: Optional[str] = None) -> Optional[bool]:
        """
        Get power status for a system that has power management configured.

        :param system: Cobbler system
        :type system: System
        :param user: power management user
        :param password: power management password
        :return: if system is powered on
        """

        return self._power(system, "status", user, password)



import pytest

from cobbler.api import CobblerAPI
from cobbler.cexceptions import CX
from cobbler.cobbler_collections.manager import CollectionManager
from cobbler.templar import Templar


@pytest.fixture(scope="function")
def setup_cheetah_macros_file():
    with open("/etc/cobbler/cheetah_macros", "w") as f:
        f.writelines(["## define Cheetah functions here and reuse them throughout your templates\n",
                      "\n",
                      "#def $myMethodInMacros($a)\n",
                      "Text in method: $a\n",
                      "#end def\n"])
    yield
    with open("/etc/cobbler/cheetah_macros", "w") as f:
        f.writelines(["## define Cheetah functions here and reuse them throughout your templates\n",
                      "\n",
                      "\n"])


def test_check_for_invalid_imports():
    # Arrange
    test_api = CobblerAPI()
    test_collection_mgr = CollectionManager(test_api)
    test_templar = Templar(test_collection_mgr)
    testdata = "#import json"

    # Act & Assert
    with pytest.raises(CX):
        test_templar.check_for_invalid_imports(testdata)


def test_render():
    # Arrange
    test_api = CobblerAPI()
    test_collection_mgr = CollectionManager(test_api)
    test_templar = Templar(test_collection_mgr)

    # Act
    result = test_templar.render("", {}, None, template_type="cheetah")

    # Assert
    assert result == ""


def test_render_cheetah():
    # Arrange
    test_api = CobblerAPI()
    test_collection_mgr = CollectionManager(test_api)
    test_templar = Templar(test_collection_mgr)

    # Act
    result = test_templar.render_cheetah("$test", {"test": 5})

    # Assert
    assert result == '5'


@pytest.mark.usefixtures("setup_cheetah_macros_file")
@pytest.mark.skip("Macros only work if we restart cobblerd")
def test_cheetah_macros():
    # Arrange
    test_api = CobblerAPI()
    test_collection_mgr = CollectionManager(test_api)
    templar = Templar(test_collection_mgr)

    # Act
    result = templar.render_cheetah("$myMethodInMacros(5)", {})

    # Assert
    assert result == "Text in method: 5\n"


def test_render_jinja2():
    # Arrange
    test_api = CobblerAPI()
    test_collection_mgr = CollectionManager(test_api)
    test_templar = Templar(test_collection_mgr)

    # Act
    result = test_templar.render_jinja2("{{ foo }}", {"foo": "Test successful"})

    # Assert
    assert result == "Test successful"



"""
Replicate from a Cobbler master.

Copyright 2007-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>
Scott Henson <shenson@redhat.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import fnmatch
import logging
import os
import xmlrpc.client
from typing import Optional

from cobbler import utils

OBJ_TYPES = ["distro", "profile", "system", "repo", "image", "mgmtclass", "package", "file"]


class Replicate:
    """
    This class contains the magic to replicate a Cobbler instance to another Cobbler instance.
    """

    def __init__(self, collection_mgr):
        """
        Constructor

        :param collection_mgr: The collection manager which holds all information available in Cobbler.
        """
        self.collection_mgr = collection_mgr
        self.settings = collection_mgr.settings()
        self.api = collection_mgr.api
        self.remote = None
        self.uri = None
        self.logger = logging.getLogger()

    def rsync_it(self, from_path, to_path, type: Optional[str] = None):
        """
        Rsync from a source to a destination with the rsync options Cobbler was configured with.

        :param from_path: The source to rsync from.
        :param to_path: The destination to rsync to.
        :param type: If set to "repo" this will take the repo rsync options instead of the global ones.
        """
        from_path = "%s::%s" % (self.master, from_path)
        if type == 'repo':
            cmd = "rsync %s %s %s" % (self.settings.replicate_repo_rsync_options, from_path, to_path)
        else:
            cmd = "rsync %s %s %s" % (self.settings.replicate_rsync_options, from_path, to_path)

        rc = utils.subprocess_call(cmd, shell=True)
        if rc != 0:
            self.logger.info("rsync failed")

    # -------------------------------------------------------

    def remove_objects_not_on_master(self, obj_type):
        """
        Remove objects on this slave which are not on the master.

        :param obj_type: The type of object which should be synchronized.
        """
        locals = utils.lod_to_dod(self.local_data[obj_type], "uid")
        remotes = utils.lod_to_dod(self.remote_data[obj_type], "uid")

        for (luid, ldata) in list(locals.items()):
            if luid not in remotes:
                try:
                    self.logger.info("removing %s %s" % (obj_type, ldata["name"]))
                    self.api.remove_item(obj_type, ldata["name"], recursive=True)
                except Exception:
                    utils.log_exc()

    # -------------------------------------------------------

    def add_objects_not_on_local(self, obj_type):
        """
        Add objects locally which are not present on the slave but on the master.

        :param obj_type:
        """
        locals = utils.lod_to_dod(self.local_data[obj_type], "uid")
        remotes = utils.lod_sort_by_key(self.remote_data[obj_type], "depth")

        for rdata in remotes:

            # do not add the system if it is not on the transfer list
            if not rdata["name"] in self.must_include[obj_type]:
                continue

            if not rdata["uid"] in locals:
                creator = getattr(self.api, "new_%s" % obj_type)
                newobj = creator()
                newobj.from_dict(utils.revert_strip_none(rdata))
                try:
                    self.logger.info("adding %s %s" % (obj_type, rdata["name"]))
                    if not self.api.add_item(obj_type, newobj):
                        self.logger.error("failed to add %s %s" % (obj_type, rdata["name"]))
                except Exception:
                    utils.log_exc()

    # -------------------------------------------------------

    def replace_objects_newer_on_remote(self, obj_type):
        """
        Replace objects which are newer on the local slave then on the remote slave

        :param obj_type: The type of object to synchronize.
        """
        locals = utils.lod_to_dod(self.local_data[obj_type], "uid")
        remotes = utils.lod_to_dod(self.remote_data[obj_type], "uid")

        for (ruid, rdata) in list(remotes.items()):
            # do not add the system if it is not on the transfer list
            if not rdata["name"] in self.must_include[obj_type]:
                continue

            if ruid in locals:
                ldata = locals[ruid]
                if ldata["mtime"] < rdata["mtime"]:

                    if ldata["name"] != rdata["name"]:
                        self.logger.info("removing %s %s" % (obj_type, ldata["name"]))
                        self.api.remove_item(obj_type, ldata["name"], recursive=True)
                    creator = getattr(self.api, "new_%s" % obj_type)
                    newobj = creator()
                    newobj.from_dict(utils.revert_strip_none(rdata))
                    try:
                        self.logger.info("updating %s %s" % (obj_type, rdata["name"]))
                        if not self.api.add_item(obj_type, newobj):
                            self.logger.error("failed to update %s %s" % (obj_type, rdata["name"]))
                    except Exception:
                        utils.log_exc()

    # -------------------------------------------------------

    def replicate_data(self):
        """
        Replicate the local and remote data to each another.

        """
        self.local_data = {}
        self.remote_data = {}
        self.remote_settings = self.remote.get_settings()

        self.logger.info("Querying Both Servers")
        for what in OBJ_TYPES:
            self.remote_data[what] = self.remote.get_items(what)
            self.local_data[what] = self.local.get_items(what)

        self.generate_include_map()

        if self.prune:
            self.logger.info("Removing Objects Not Stored On Master")
            obj_types = OBJ_TYPES[:]
            if len(self.system_patterns) == 0 and "system" in obj_types:
                obj_types.remove("system")
            for what in obj_types:
                self.remove_objects_not_on_master(what)
        else:
            self.logger.info("*NOT* Removing Objects Not Stored On Master")

        if not self.omit_data:
            self.logger.info("Rsyncing distros")
            for distro in list(self.must_include["distro"].keys()):
                if self.must_include["distro"][distro] == 1:
                    self.logger.info("Rsyncing distro %s" % distro)
                    target = self.remote.get_distro(distro)
                    target_webdir = os.path.join(self.remote_settings["webdir"], "distro_mirror")
                    tail = utils.path_tail(target_webdir, target["kernel"])
                    if tail != "":
                        try:
                            # path_tail(a,b) returns something that looks like
                            # an absolute path, but it's really the sub-path
                            # from a that is contained in b. That means we want
                            # the first element of the path
                            dest = os.path.join(self.settings.webdir, "distro_mirror", tail.split("/")[1])
                            self.rsync_it("distro-%s" % target["name"], dest)
                        except:
                            self.logger.error("Failed to rsync distro %s" % distro)
                            continue
                    else:
                        self.logger.warning("Skipping distro %s, as it doesn't appear to live under distro_mirror"
                                            % distro)

            self.logger.info("Rsyncing repos")
            for repo in list(self.must_include["repo"].keys()):
                if self.must_include["repo"][repo] == 1:
                    self.rsync_it("repo-%s" % repo, os.path.join(self.settings.webdir, "repo_mirror", repo), "repo")

            self.logger.info("Rsyncing distro repo configs")
            self.rsync_it("cobbler-distros/config/", os.path.join(self.settings.webdir, "distro_mirror", "config"))
            self.logger.info("Rsyncing automatic installation templates & snippets")
            self.rsync_it("cobbler-templates", self.settings.autoinstall_templates_dir)
            self.rsync_it("cobbler-snippets", self.settings.autoinstall_snippets_dir)
            self.logger.info("Rsyncing triggers")
            self.rsync_it("cobbler-triggers", "/var/lib/cobbler/triggers")
            self.logger.info("Rsyncing scripts")
            self.rsync_it("cobbler-scripts", "/var/lib/cobbler/scripts")
        else:
            self.logger.info("*NOT* Rsyncing Data")

        self.logger.info("Adding Objects Not Stored On Local")
        for what in OBJ_TYPES:
            self.add_objects_not_on_local(what)

        self.logger.info("Updating Objects Newer On Remote")
        for what in OBJ_TYPES:
            self.replace_objects_newer_on_remote(what)

    def link_distros(self):
        """
        Link a distro from its location into the web directory to make it available for usage.
        """
        for distro in self.api.distros():
            self.logger.debug("Linking Distro %s" % distro.name)
            utils.link_distro(self.settings, distro)

    def generate_include_map(self):
        """
        Not known what this exactly does.
        """
        self.remote_names = {}
        self.remote_dict = {}
        self.must_include = {
            "distro": {},
            "profile": {},
            "system": {},
            "image": {},
            "repo": {},
            "mgmtclass": {},
            "package": {},
            "file": {}
        }

        for ot in OBJ_TYPES:
            self.remote_names[ot] = list(utils.lod_to_dod(self.remote_data[ot], "name").keys())
            self.remote_dict[ot] = utils.lod_to_dod(self.remote_data[ot], "name")
            if self.sync_all:
                for names in self.remote_dict[ot]:
                    self.must_include[ot][names] = 1

        self.logger.debug("remote names struct is %s" % self.remote_names)

        if not self.sync_all:
            # include all profiles that are matched by a pattern
            for obj_type in OBJ_TYPES:
                patvar = getattr(self, "%s_patterns" % obj_type)
                self.logger.debug("* Finding Explicit %s Matches" % obj_type)
                for pat in patvar:
                    for remote in self.remote_names[obj_type]:
                        self.logger.debug("?: seeing if %s looks like %s" % (remote, pat))
                        if fnmatch.fnmatch(remote, pat):
                            self.logger.debug("Adding %s for pattern match %s." % (remote, pat))
                            self.must_include[obj_type][remote] = 1

            # include all profiles that systems require whether they are explicitly included or not
            self.logger.debug("* Adding Profiles Required By Systems")
            for sys in list(self.must_include["system"].keys()):
                pro = self.remote_dict["system"][sys].get("profile", "")
                self.logger.debug("?: system %s requires profile %s." % (sys, pro))
                if pro != "":
                    self.logger.debug("Adding profile %s for system %s." % (pro, sys))
                    self.must_include["profile"][pro] = 1

            # include all profiles that subprofiles require whether they are explicitly included or not very deep
            # nesting is possible
            self.logger.debug("* Adding Profiles Required By SubProfiles")
            while True:
                loop_exit = True
                for pro in list(self.must_include["profile"].keys()):
                    parent = self.remote_dict["profile"][pro].get("parent", "")
                    if parent != "":
                        if parent not in self.must_include["profile"]:
                            self.logger.debug("Adding parent profile %s for profile %s." % (parent, pro))
                            self.must_include["profile"][parent] = 1
                            loop_exit = False
                if loop_exit:
                    break

            # require all distros that any profiles in the generated list requires whether they are explicitly included
            # or not
            self.logger.debug("* Adding Distros Required By Profiles")
            for p in list(self.must_include["profile"].keys()):
                distro = self.remote_dict["profile"][p].get("distro", "")
                if not distro == "<<inherit>>" and not distro == "~":
                    self.logger.debug("Adding distro %s for profile %s." % (distro, p))
                    self.must_include["distro"][distro] = 1

            # require any repos that any profiles in the generated list requires whether they are explicitly included
            # or not
            self.logger.debug("* Adding Repos Required By Profiles")
            for p in list(self.must_include["profile"].keys()):
                repos = self.remote_dict["profile"][p].get("repos", [])
                if repos != "<<inherit>>":
                    for r in repos:
                        self.logger.debug("Adding repo %s for profile %s." % (r, p))
                        self.must_include["repo"][r] = 1

            # include all images that systems require whether they are explicitly included or not
            self.logger.debug("* Adding Images Required By Systems")
            for sys in list(self.must_include["system"].keys()):
                img = self.remote_dict["system"][sys].get("image", "")
                self.logger.debug("?: system %s requires image %s." % (sys, img))
                if img != "":
                    self.logger.debug("Adding image %s for system %s." % (img, sys))
                    self.must_include["image"][img] = 1

    # -------------------------------------------------------

    def run(self, cobbler_master=None, port: str = "80", distro_patterns=None, profile_patterns=None,
            system_patterns=None, repo_patterns=None, image_patterns=None, mgmtclass_patterns=None,
            package_patterns=None, file_patterns=None, prune: bool = False, omit_data=False, sync_all: bool = False,
            use_ssl: bool = False):
        """
        Get remote profiles and distros and sync them locally

        :param cobbler_master: The remote url of the master server.
        :param port: The remote port of the master server.
        :param distro_patterns: The pattern of distros to sync.
        :param profile_patterns: The pattern of profiles to sync.
        :param system_patterns: The pattern of systems to sync.
        :param repo_patterns: The pattern of repositories to sync.
        :param image_patterns: The pattern of images to sync.
        :param mgmtclass_patterns: The pattern of management classes to sync.
        :param package_patterns: The pattern of packages to sync.
        :param file_patterns: The pattern of files to sync.
        :param prune: If the local server should be pruned before coping stuff.
        :param omit_data: If the data behind images etc should be omitted or not.
        :param sync_all: If everything should be synced (then the patterns are useless) or not.
        :param use_ssl: If HTTPS or HTTP should be used.
        """

        self.port = str(port)
        self.distro_patterns = distro_patterns.split()
        self.profile_patterns = profile_patterns.split()
        self.system_patterns = system_patterns.split()
        self.repo_patterns = repo_patterns.split()
        self.image_patterns = image_patterns.split()
        self.mgmtclass_patterns = mgmtclass_patterns.split()
        self.package_patterns = package_patterns.split()
        self.file_patterns = file_patterns.split()
        self.omit_data = omit_data
        self.prune = prune
        self.sync_all = sync_all
        self.use_ssl = use_ssl

        if self.use_ssl:
            protocol = 'https'
        else:
            protocol = 'http'

        if cobbler_master is not None:
            self.master = cobbler_master
        elif len(self.settings.cobbler_master) > 0:
            self.master = self.settings.cobbler_master
        else:
            utils.die('No Cobbler master specified, try --master.')

        self.uri = '%s://%s:%s/cobbler_api' % (protocol, self.master, self.port)

        self.logger.info("cobbler_master      = %s" % cobbler_master)
        self.logger.info("port                = %s" % self.port)
        self.logger.info("distro_patterns     = %s" % self.distro_patterns)
        self.logger.info("profile_patterns    = %s" % self.profile_patterns)
        self.logger.info("system_patterns     = %s" % self.system_patterns)
        self.logger.info("repo_patterns       = %s" % self.repo_patterns)
        self.logger.info("image_patterns      = %s" % self.image_patterns)
        self.logger.info("mgmtclass_patterns  = %s" % self.mgmtclass_patterns)
        self.logger.info("package_patterns    = %s" % self.package_patterns)
        self.logger.info("file_patterns       = %s" % self.file_patterns)
        self.logger.info("omit_data           = %s" % self.omit_data)
        self.logger.info("sync_all            = %s" % self.sync_all)
        self.logger.info("use_ssl             = %s" % self.use_ssl)

        self.logger.info("XMLRPC endpoint: %s" % self.uri)
        self.logger.debug("test ALPHA")
        self.remote = xmlrpc.client.Server(self.uri)
        self.logger.debug("test BETA")
        self.remote.ping()
        self.local = xmlrpc.client.Server("http://127.0.0.1:%s/cobbler_api" % self.settings.http_port)
        self.local.ping()

        self.replicate_data()
        self.link_distros()
        self.logger.info("Syncing")
        self.api.sync()
        self.logger.info("Done")



"""
Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import logging
import os
import pathlib
import random
import tempfile
import threading
from configparser import ConfigParser
from pathlib import Path
from typing import Dict, List, Optional, Union

from cobbler.actions import status, hardlink, sync, buildiso, replicate, report, log, acl, check, reposync, grubimage
from cobbler import autoinstall_manager, autoinstallgen, download_manager, enums, module_loader, power_manager
from cobbler import settings, tftpgen, utils, yumgen
from cobbler.cobbler_collections import manager
from cobbler.items import distro, file, image, menu, mgmtclass, package, profile, repo, system

# FIXME: add --quiet depending on if not --verbose?
RSYNC_CMD = "rsync -a %s '%s' %s --progress"


# notes on locking:
# - CobblerAPI is a singleton object
# - The XMLRPC variants allow 1 simultaneous request, therefore we flock on our settings file for now on a request by
#   request basis.


class CobblerAPI:
    """
    Python API module for Cobbler.
    See source for cobbler.py, or pydoc, for example usage.
    Cli apps and daemons should import api.py, and no other Cobbler code.
    """
    __shared_state = {}
    __has_loaded = False

    def __init__(self, is_cobblerd: bool = False, settingsfile_location: str = "/etc/cobbler/settings.yaml",
                 execute_settings_automigration: bool = True):
        """
        Constructor

        :param is_cobblerd: Wether this API is run as a daemon or not.
        :param settingsfile_location: The location of the settings file on the disk.
        """

        # FIXME: this should be switchable through some simple system

        self.__dict__ = CobblerAPI.__shared_state
        self.perms_ok = False
        if not CobblerAPI.__has_loaded:
            # NOTE: we do not log all API actions, because a simple CLI invocation may call adds and such to load the
            # config, which would just fill up the logs, so we'll do that logging at CLI level (and remote.py web
            # service level) instead.

            random.seed()
            self.is_cobblerd = is_cobblerd
            if is_cobblerd:
                main_thread = threading.main_thread()
                main_thread.setName("Daemon")

            self.logger = logging.getLogger()

            # FIXME: consolidate into 1 server instance

            self.selinux_enabled = utils.is_selinux_enabled()
            self.dist, self.os_version = utils.os_release()
            self._settings = self.__generate_settings(pathlib.Path(settingsfile_location),
                                                      execute_settings_automigration)

            CobblerAPI.__has_loaded = True

            # load the modules first, or nothing else works...
            module_loader.load_modules()

            # import signatures
            try:
                utils.load_signatures(self.settings().signature_path)
            except Exception as e:
                self.log("Failed to load signatures from %s: %s" % (self.settings().signature_path, e))
                return

            self._collection_mgr = manager.CollectionManager(self)
            self.deserialize()

            self.log("%d breeds and %d OS versions read from the signature file"
                     % (len(utils.get_valid_breeds()), len(utils.get_valid_os_versions())))

            self.authn = self.get_module_from_file(
                "authentication",
                "module",
                "authn_configfile"
            )
            self.authz = self.get_module_from_file(
                "authorization",
                "module",
                "authz_allowall"
            )

            # FIXME: pass more loggers around, and also see that those using things via tasks construct their own
            #  yumgen/tftpgen versus reusing this one, which has the wrong logger (most likely) for background tasks.

            self.autoinstallgen = autoinstallgen.AutoInstallationGen(self._collection_mgr)
            self.yumgen = yumgen.YumGen(self._collection_mgr)
            self.tftpgen = tftpgen.TFTPGen(self._collection_mgr)
            self.power_mgr = power_manager.PowerManager(self, self._collection_mgr)
            self.logger.debug("API handle initialized")
            self.perms_ok = True

    def __generate_settings(self, settings_path: Path,
                            execute_settings_automigration: bool = True) -> settings.Settings:
        # Read in YAML file and get dict
        yaml_dict = settings.read_yaml_file(settings_path)

        # Add or update auto migrate settings value
        yaml_dict["auto_migrate_settings"] = execute_settings_automigration

        # Take dict and use it in migrations
        migrated_settings = settings.migrate(yaml_dict, settings_path)

        # After the migration is done, save result in the settings object
        new_settings = settings.Settings()
        new_settings.from_dict(migrated_settings)

        # save to disk
        new_settings.save(settings_path)

        # Return object
        return new_settings

    # ==========================================================

    def is_selinux_enabled(self) -> bool:
        """
        Returns whether selinux is enabled on the Cobbler server.
        We check this just once at Cobbler API init time, because a restart is required to change this; this does
        /not/ check enforce/permissive, nor does it need to.
        """
        return self.selinux_enabled

    def is_selinux_supported(self) -> bool:
        """
        Returns whether or not the OS is sufficient enough to run with SELinux enabled (currently EL 5 or later).

        :returns: False per default. If Distro is Redhat and Version >= 5 then it returns true.
        """
        # FIXME: This detection is flawed. There is more than just Rhel with selinux and the original implementation was
        #        too broad.
        if ("red hat" in self.dist or "redhat" in self.dist) and self.os_version >= 5:
            return True
        # doesn't support public_content_t
        return False

    # ==========================================================

    def last_modified_time(self) -> float:
        """
        Returns the time of the last modification to Cobbler, made by any API instance, regardless of the serializer
        type.

        :returns: 0 if there is no file where the information required for this method is saved.
        """
        # FIXME: This fails in case the file required is not available
        if not os.path.exists("/var/lib/cobbler/.mtime"):
            fd = open("/var/lib/cobbler/.mtime", 'w')
            fd.write("0")
            fd.close()
            return float(0)
        fd = open("/var/lib/cobbler/.mtime", 'r')
        data = fd.read().strip()
        return float(data)

    # ==========================================================

    def log(self, msg: str, args=None, debug: bool = False):
        """
        Logs a message with the already initiated logger of this object.

        :param msg: The message to log.
        :param args: Optional message which gets appended to the main msg with a ';'.
        :param debug: Weather the logged message is a debug message (true) or info (false).

        .. deprecated:: 3.3.0
           We should use the standard logger.
        """
        if debug:
            logger = self.logger.debug
        else:
            logger = self.logger.info
        if args is None:
            logger("%s" % msg)
        else:
            logger("%s; %s" % (msg, str(args)))

    # ==========================================================

    def version(self, extended: bool = False) -> Union[float, Dict[str, Union[str, list]]]:
        """
        What version is Cobbler?

        If extended == False, returns a float for backwards compatibility
        If extended == True, returns a dict:

            gitstamp      -- the last git commit hash
            gitdate       -- the last git commit date on the builder machine
            builddate     -- the time of the build
            version       -- something like "1.3.2"
            version_tuple -- something like [ 1, 3, 2 ]

        :param extended: False returns a float, True a Dictionary.
        """
        config = ConfigParser()
        config.read("/etc/cobbler/version")
        data: Dict[str, Union[str, list]] = {
            "gitdate": config.get("cobbler", "gitdate"),
            "gitstamp": config.get("cobbler", "gitstamp"),
            "builddate": config.get("cobbler", "builddate"),
            "version": config.get("cobbler", "version"),
            "version_tuple": []
        }
        # dont actually read the version_tuple from the version file
        for num in data["version"].split("."):
            data["version_tuple"].append(int(num))

        if not extended:
            # for backwards compatibility and use with koan's comparisons
            elems = data["version_tuple"]
            return int(elems[0]) + 0.1 * int(elems[1]) + 0.001 * int(elems[2])
        else:
            return data

    # ==========================================================

    def get_item(self, what: str, name: str):
        """
        Get a general item.

        :param what: The item type to retrieve from the internal database.
        :param name: The name of the item to retrieve.
        :return: An item of the desired type.
        """
        # self.log("get_item", [what, name], debug=True)
        item = self._collection_mgr.get_items(what).get(name)
        return item

    def get_items(self, what: str):
        """
        Get all items of a collection.

        :param what: The collection to query.
        :return: The items which were queried. May return no items.
        """
        # self.log("get_items", [what], debug=True)
        items = self._collection_mgr.get_items(what)
        return items

    def distros(self):
        """
        Return the current list of distributions
        """
        return self.get_items("distro")

    def profiles(self):
        """
        Return the current list of profiles
        """
        return self.get_items("profile")

    def systems(self):
        """
        Return the current list of systems
        """
        return self.get_items("system")

    def repos(self):
        """
        Return the current list of repos
        """
        return self.get_items("repo")

    def images(self):
        """
        Return the current list of images
        """
        return self.get_items("image")

    def settings(self):
        """
        Return the application configuration
        """
        return self._settings

    def mgmtclasses(self):
        """
        Return the current list of mgmtclasses
        """
        return self.get_items("mgmtclass")

    def packages(self):
        """
        Return the current list of packages
        """
        return self.get_items("package")

    def files(self):
        """
        Return the current list of files
        """
        return self.get_items("file")

    def menus(self):
        """
        Return the current list of menus
        """
        return self.get_items("menu")

    # =======================================================================

    def copy_item(self, what: str, ref, newname: str):
        """
        General copy method which is called by the specific methods.

        :param what: The collection type which gets copied.
        :param ref: The object itself which gets copied.
        :param newname: The new name of the newly created object.
        """
        self.log("copy_item(%s)" % what, [ref.name, newname])
        self.get_items(what).copy(ref, newname)

    def copy_distro(self, ref, newname: str):
        """
        This method copies a distro which is just different in the name of the object.

        :param ref: The object itself which gets copied.
        :param newname: The new name of the newly created object.
        """
        self.copy_item("distro", ref, newname)

    def copy_profile(self, ref, newname: str):
        """
        This method copies a profile which is just different in the name of the object.

        :param ref: The object itself which gets copied.
        :param newname: The new name of the newly created object.
        """
        self.copy_item("profile", ref, newname)

    def copy_system(self, ref, newname: str):
        """
        This method copies a system which is just different in the name of the object.

        :param ref: The object itself which gets copied.
        :param newname: The new name of the newly created object.
        """
        self.copy_item("system", ref, newname)

    def copy_repo(self, ref, newname: str):
        """
        This method copies a repository which is just different in the name of the object.

        :param ref: The object itself which gets copied.
        :param newname: The new name of the newly created object.
        """
        self.copy_item("repo", ref, newname)

    def copy_image(self, ref, newname: str):
        """
        This method copies an image which is just different in the name of the object.

        :param ref: The object itself which gets copied.
        :param newname: The new name of the newly created object.
        """
        self.copy_item("image", ref, newname)

    def copy_mgmtclass(self, ref, newname: str):
        """
        This method copies a management class which is just different in the name of the object.

        :param ref: The object itself which gets copied.
        :param newname: The new name of the newly created object.
        """
        self.copy_item("mgmtclass", ref, newname)

    def copy_package(self, ref, newname: str):
        """
        This method copies a package which is just different in the name of the object.

        :param ref: The object itself which gets copied.
        :param newname: The new name of the newly created object.
        """
        self.copy_item("package", ref, newname)

    def copy_file(self, ref, newname: str):
        """
        This method copies a file which is just different in the name of the object.

        :param ref: The object itself which gets copied.
        :param newname: The new name of the newly created object.
        """
        self.copy_item("file", ref, newname)

    def copy_menu(self, ref, newname: str):
        """
        This method copies a file which is just different in the name of the object.

        :param ref: The object itself which gets copied.
        :param newname: The new name of the newly created object.
        """
        self.copy_item("menu", ref, newname)

    # ==========================================================================

    def remove_item(self, what: str, ref, recursive: bool = False, delete: bool = True, with_triggers: bool = True):
        """
        Remove a general item. This method should not be used by an external api. Please use the specific
        remove_<itemtype> methods.

        :param what: The type of the item.
        :param ref: The internal unique handle for the item.
        :param recursive: If the item should recursively should delete dependencies on itself.
        :param delete: Not known what this parameter does exactly.
        :param with_triggers: Whether you would like to have the removal triggers executed or not.
        """
        if isinstance(what, str):
            if isinstance(ref, str):
                ref = self.get_item(what, ref)
                if ref is None:
                    return  # nothing to remove
        self.log("remove_item(%s)" % what, [ref.name])
        self.get_items(what).remove(ref.name, recursive=recursive, with_delete=delete, with_triggers=with_triggers)

    def remove_distro(self, ref, recursive: bool = False, delete: bool = True, with_triggers: bool = True):
        """
        Remove a distribution from Cobbler.

        :param ref: The internal unique handle for the item.
        :param recursive: If the item should recursively should delete dependencies on itself.
        :param delete: Not known what this parameter does exactly.
        :param with_triggers: Whether you would like to have the removal triggers executed or not.
        """
        self.remove_item("distro", ref, recursive=recursive, delete=delete, with_triggers=with_triggers)

    def remove_profile(self, ref, recursive: bool = False, delete: bool = True, with_triggers: bool = True):
        """
        Remove a profile from Cobbler.

        :param ref: The internal unique handle for the item.
        :param recursive: If the item should recursively should delete dependencies on itself.
        :param delete: Not known what this parameter does exactly.
        :param with_triggers: Whether you would like to have the removal triggers executed or not.
        """
        self.remove_item("profile", ref, recursive=recursive, delete=delete, with_triggers=with_triggers)

    def remove_system(self, ref, recursive: bool = False, delete: bool = True, with_triggers: bool = True):
        """
        Remove a system from Cobbler.

        :param ref: The internal unique handle for the item.
        :param recursive: If the item should recursively should delete dependencies on itself.
        :param delete: Not known what this parameter does exactly.
        :param with_triggers: Whether you would like to have the removal triggers executed or not.
        """
        self.remove_item("system", ref, recursive=recursive, delete=delete, with_triggers=with_triggers)

    def remove_repo(self, ref, recursive: bool = False, delete: bool = True, with_triggers: bool = True):
        """
        Remove a repository from Cobbler.

        :param ref: The internal unique handle for the item.
        :param recursive: If the item should recursively should delete dependencies on itself.
        :param delete: Not known what this parameter does exactly.
        :param with_triggers: Whether you would like to have the removal triggers executed or not.
        """
        self.remove_item("repo", ref, recursive=recursive, delete=delete, with_triggers=with_triggers)

    def remove_image(self, ref, recursive: bool = False, delete: bool = True, with_triggers: bool = True):
        """
        Remove a image from Cobbler.

        :param ref: The internal unique handle for the item.
        :param recursive: If the item should recursively should delete dependencies on itself.
        :param delete: Not known what this parameter does exactly.
        :param with_triggers: Whether you would like to have the removal triggers executed or not.
        """
        self.remove_item("image", ref, recursive=recursive, delete=delete, with_triggers=with_triggers)

    def remove_mgmtclass(self, ref, recursive: bool = False, delete: bool = True, with_triggers: bool = True):
        """
        Remove a management class from Cobbler.

        :param ref: The internal unique handle for the item.
        :param recursive: If the item should recursively should delete dependencies on itself.
        :param delete: Not known what this parameter does exactly.
        :param with_triggers: Whether you would like to have the removal triggers executed or not.
        """
        self.remove_item("mgmtclass", ref, recursive=recursive, delete=delete, with_triggers=with_triggers)

    def remove_package(self, ref, recursive: bool = False, delete: bool = True, with_triggers: bool = True):
        """
        Remove a package from Cobbler.

        :param ref: The internal unique handle for the item.
        :param recursive: If the item should recursively should delete dependencies on itself.
        :param delete: Not known what this parameter does exactly.
        :param with_triggers: Whether you would like to have the removal triggers executed or not.
        """
        self.remove_item("package", ref, recursive=recursive, delete=delete, with_triggers=with_triggers)

    def remove_file(self, ref, recursive: bool = False, delete: bool = True, with_triggers: bool = True):
        """
        Remove a file from Cobbler.

        :param ref: The internal unique handle for the item.
        :param recursive: If the item should recursively should delete dependencies on itself.
        :param delete: Not known what this parameter does exactly.
        :param with_triggers: Whether you would like to have the removal triggers executed or not.
        """
        self.remove_item("file", ref, recursive=recursive, delete=delete, with_triggers=with_triggers)

    def remove_menu(self, ref, recursive=False, delete=True, with_triggers=True):
        """
        Remove a menu from Cobbler.

        :param ref: The internal unique handle for the item.
        :param recursive: If the item should recursively should delete dependencies on itself.
        :param delete: Not known what this parameter does exactly.
        :param with_triggers: Whether you would like to have the removal triggers executed or not.
        """
        self.remove_item("menu", ref, recursive=recursive, delete=delete, with_triggers=with_triggers)

    # ==========================================================================

    def rename_item(self, what: str, ref, newname: str):
        """
        Remove a general item. This method should not be used by an external api. Please use the specific
        rename_<itemtype> methods.

        :param what: The type of object which should be renamed.
        :param ref: The internal unique handle for the item.
        :param newname: The new name for the item.
        """
        self.log("rename_item(%s)" % what, [ref.name, newname])
        self.get_items(what).rename(ref, newname)

    def rename_distro(self, ref, newname: str):
        """
        Rename a distro to a new name.

        :param ref: The internal unique handle for the item.
        :param newname: The new name for the item.
        """
        self.rename_item("distro", ref, newname)

    def rename_profile(self, ref, newname: str):
        """
        Rename a profile to a new name.

        :param ref: The internal unique handle for the item.
        :param newname: The new name for the item.
        """
        self.rename_item("profile", ref, newname)

    def rename_system(self, ref, newname: str):
        """
        Rename a system to a new name.

        :param ref: The internal unique handle for the item.
        :param newname: The new name for the item.
        """
        self.rename_item("system", ref, newname)

    def rename_repo(self, ref, newname: str):
        """
        Rename a repository to a new name.

        :param ref: The internal unique handle for the item.
        :param newname: The new name for the item.
        """
        self.rename_item("repo", ref, newname)

    def rename_image(self, ref, newname: str):
        """
        Rename an image to a new name.

        :param ref: The internal unique handle for the item.
        :param newname: The new name for the item.
        """
        self.rename_item("image", ref, newname)

    def rename_mgmtclass(self, ref, newname: str):
        """
        Rename a management class to a new name.

        :param ref: The internal unique handle for the item.
        :param newname: The new name for the item.
        """
        self.rename_item("mgmtclass", ref, newname)

    def rename_package(self, ref, newname: str):
        """
        Rename a package to a new name.

        :param ref: The internal unique handle for the item.
        :param newname: The new name for the item.
        """
        self.rename_item("package", ref, newname)

    def rename_file(self, ref, newname: str):
        """
        Rename a file to a new name.

        :param ref: The internal unique handle for the item.
        :param newname: The new name for the item.
        """
        self.rename_item("file", ref, newname)

    def rename_menu(self, ref, newname: str):
        """
        Rename a menu to a new name.

        :param ref: The internal unique handle for the item.
        :param newname: The new name for the item.
        """
        self.rename_item("menu", ref, newname)

    # ==========================================================================

    # FIXME: add a new_item method

    def new_distro(self, is_subobject: bool = False):
        """
        Returns a new empty distro object. This distro is not automatically persisted. Persistance is achieved via
        ``save()``.

        :param is_subobject: If the object created is a subobject or not.
        :return: An empty Distro object.
        """
        self.log("new_distro", [is_subobject])
        return distro.Distro(self._collection_mgr, is_subobject=is_subobject)

    def new_profile(self, is_subobject: bool = False):
        """
        Returns a new empty profile object. This profile is not automatically persisted. Persistence is achieved via
        ``save()``.

        :param is_subobject: If the object created is a subobject or not.
        :return: An empty Profile object.
        """
        self.log("new_profile", [is_subobject])
        return profile.Profile(self._collection_mgr, is_subobject=is_subobject)

    def new_system(self, is_subobject: bool = False):
        """
        Returns a new empty system object. This system is not automatically persisted. Persistence is achieved via
        ``save()``.

        :param is_subobject: If the object created is a subobject or not.
        :return: An empty System object.
        """
        self.log("new_system", [is_subobject])
        return system.System(self._collection_mgr, is_subobject=is_subobject)

    def new_repo(self, is_subobject: bool = False):
        """
        Returns a new empty repo object. This repository is not automatically persisted. Persistence is achieved via
        ``save()``.

        :param is_subobject: If the object created is a subobject or not.
        :return: An empty repo object.
        """
        self.log("new_repo", [is_subobject])
        return repo.Repo(self._collection_mgr, is_subobject=is_subobject)

    def new_image(self, is_subobject: bool = False):
        """
        Returns a new empty image object. This image is not automatically persisted. Persistence is achieved via
        ``save()``.

        :param is_subobject: If the object created is a subobject or not.
        :return: An empty image object.
        """
        self.log("new_image", [is_subobject])
        return image.Image(self._collection_mgr, is_subobject=is_subobject)

    def new_mgmtclass(self, is_subobject: bool = False):
        """
        Returns a new empty mgmtclass object. This mgmtclass is not automatically persisted. Persistence is achieved via
        ``save()``.

        :param is_subobject: If the object created is a subobject or not.
        :return: An empty mgmtclass object.
        """
        self.log("new_mgmtclass", [is_subobject])
        return mgmtclass.Mgmtclass(self._collection_mgr, is_subobject=is_subobject)

    def new_package(self, is_subobject: bool = False):
        """
        Returns a new empty package object. This package is not automatically persisted. Persistence is achieved via
        ``save()``.

        :param is_subobject: If the object created is a subobject or not.
        :return: An empty Package object.
        """
        self.log("new_package", [is_subobject])
        return package.Package(self._collection_mgr, is_subobject=is_subobject)

    def new_file(self, is_subobject: bool = False):
        """
        Returns a new empty file object. This file is not automatically persisted. Persistence is achieved via
        ``save()``.

        :param is_subobject: If the object created is a subobject or not.
        :return: An empty File object.
        """
        self.log("new_file", [is_subobject])
        return file.File(self._collection_mgr, is_subobject=is_subobject)

    def new_menu(self, is_subobject: bool = False):
        """
        Returns a new empty menu object. This file is not automatically persisted. Persistence is achieved via
        ``save()``.

        :param is_subobject: If the object created is a subobject or not.
        :return: An empty File object.
        """
        self.log("new_menu", [is_subobject])
        return menu.Menu(self._collection_mgr, is_subobject=is_subobject)

    # ==========================================================================

    def add_item(self, what: str, ref, check_for_duplicate_names: bool = False, save: bool = True):
        """
        Add an abstract item to a collection of its specific items. This is not meant for external use. Please reefer
        to one of the specific methods ``add_<type>``.

        :param what: The item type.
        :param ref: The identifier for the object to add to a collection.
        :param check_for_duplicate_names: If the name should be unique or can be present multiple times.
        :param save: If the item should be persisted.
        """
        self.log("add_item(%s)" % what, [ref.name])
        self.get_items(what).add(ref, check_for_duplicate_names=check_for_duplicate_names, save=save)

    def add_distro(self, ref, check_for_duplicate_names: bool = False, save: bool = True):
        """
        Add a distribution to Cobbler.

        :param ref: The identifier for the object to add to a collection.
        :param check_for_duplicate_names: If the name should be unique or can be present multiple times.
        :param save: If the item should be persisted.
        """
        self.add_item("distro", ref, check_for_duplicate_names=check_for_duplicate_names, save=save)

    def add_profile(self, ref, check_for_duplicate_names: bool = False, save: bool = True):
        """
        Add a profile to Cobbler.

        :param ref: The identifier for the object to add to a collection.
        :param check_for_duplicate_names: If the name should be unique or can be present multiple times.
        :param save: If the item should be persisted.
        """
        self.add_item("profile", ref, check_for_duplicate_names=check_for_duplicate_names, save=save)

    def add_system(self, ref, check_for_duplicate_names: bool = False, check_for_duplicate_netinfo=False,
                   save: bool = True):
        """
        Add a system to Cobbler.

        :param ref: The identifier for the object to add to a collection.
        :param check_for_duplicate_names: If the name should be unique or can be present multiple times.
        :param check_for_duplicate_netinfo: If the name of the network interface should be unique or can be present
                                            multiple times.
        :param save: If the item should be persisted.
        """
        self.add_item("system", ref, check_for_duplicate_names=check_for_duplicate_names, save=save)

    def add_repo(self, ref, check_for_duplicate_names: bool = False, save: bool = True):
        """
        Add a repository to Cobbler.

        :param ref: The identifier for the object to add to a collection.
        :param check_for_duplicate_names: If the name should be unique or can be present multiple times.
        :param save: If the item should be persisted.
        """
        self.add_item("repo", ref, check_for_duplicate_names=check_for_duplicate_names, save=save)

    def add_image(self, ref, check_for_duplicate_names: bool = False, save: bool = True):
        """
        Add an image to Cobbler.

        :param ref: The identifier for the object to add to a collection.
        :param check_for_duplicate_names: If the name should be unique or can be present multiple times.
        :param save: If the item should be persisted.
        """
        self.add_item("image", ref, check_for_duplicate_names=check_for_duplicate_names, save=save)

    def add_mgmtclass(self, ref, check_for_duplicate_names: bool = False, save: bool = True):
        """
        Add a management class to Cobbler.

        :param ref: The identifier for the object to add to a collection.
        :param check_for_duplicate_names: If the name should be unique or can be present multiple times.
        :param save: If the item should be persisted.
        """
        self.add_item("mgmtclass", ref, check_for_duplicate_names=check_for_duplicate_names, save=save)

    def add_package(self, ref, check_for_duplicate_names: bool = False, save: bool = True):
        """
        Add a package to Cobbler.

        :param ref: The identifier for the object to add to a collection.
        :param check_for_duplicate_names: If the name should be unique or can be present multiple times.
        :param save: If the item should be persisted.
        """
        self.add_item("package", ref, check_for_duplicate_names=check_for_duplicate_names, save=save)

    def add_file(self, ref, check_for_duplicate_names: bool = False, save: bool = True):
        """
        Add a file to Cobbler.

        :param ref: The identifier for the object to add to a collection.
        :param check_for_duplicate_names: If the name should be unique or can be present multiple times.
        :param save: If the item should be persisted.
        """
        self.add_item("file", ref, check_for_duplicate_names=check_for_duplicate_names, save=save)

    def add_menu(self, ref, check_for_duplicate_names=False, save=True):
        """
        Add a submenu to Cobbler.

        :param ref: The identifier for the object to add to a collection.
        :param check_for_duplicate_names: If the name should be unique or can be present multiple times.
        :param save: If the item should be persisted.
        """
        self.add_item("menu", ref, check_for_duplicate_names=check_for_duplicate_names, save=save)

    # ==========================================================================

    def find_items(self, what: str, criteria: dict = None, name: str = "", return_list: bool = True,
                   no_errors: bool = False):
        """
        This is the abstract base method for finding object int the api. It should not be used by external resources.
        Please reefer to the specific implementations of this method called ``find_<object type>``.

        :param what: The object type of the item to search for.
        :param criteria: The dictionary with the key-value pairs to find objects with.
        :param name: The name of the object.
        :param return_list: If only the first result or all results should be returned.
        :param no_errors: Silence some errors which would raise if this turned to False.
        :return: The list of items witch match the search criteria.
        """
        # self.log("find_items", [what])
        if criteria is None:
            criteria = {}

        if what == "" and ("name" in criteria or name is not None):
            return self.__find_by_name(criteria.get("name", name))

        if what not in ["distro", "profile", "system", "repo", "image", "mgmtclass", "package", "file", "menu"]:
            raise ValueError("what needs to be a valid collection!")

        items = self._collection_mgr.get_items(what)
        return items.find(name=name, return_list=return_list, no_errors=no_errors, **criteria)

    def __find_by_name(self, name: str):
        """
        This is a magic method which just searches all collections for the specified name directly,
        :param name: The name of the item(s).
        :return: The found item or None.
        """
        if not isinstance(name, str):
            raise TypeError("name of an object must be of type str!")
        collections = ["distro", "profile", "system", "repo", "image", "mgmtclass", "package", "file", "menu"]
        for collection_name in collections:
            match = self.find_items(collection_name, name=name, return_list=False)
            if match is not None:
                return match
        return None

    def find_distro(self, name: str = "", return_list=False, no_errors=False, **kargs):
        """
        Find a distribution via a name or keys specified in the ``**kargs``.

        :param name: The name to search for.
        :param return_list: If only the first result or all results should be returned.
        :param no_errors: Silence some errors which would raise if this turned to False.
        :param kargs: Additional key-value pairs which may help in finding the desired objects.
        :return: A single object or a list of all search results.
        """
        return self._collection_mgr.distros().find(name=name, return_list=return_list, no_errors=no_errors, **kargs)

    def find_profile(self, name: str = "", return_list=False, no_errors=False, **kargs):
        """
        Find a profile via a name or keys specified in the ``**kargs``.

        :param name: The name to search for.
        :param return_list: If only the first result or all results should be returned.
        :param no_errors: Silence some errors which would raise if this turned to False.
        :param kargs: Additional key-value pairs which may help in finding the desired objects.
        :return: A single object or a list of all search results.
        """
        return self._collection_mgr.profiles().find(name=name, return_list=return_list, no_errors=no_errors, **kargs)

    def find_system(self, name: str = "", return_list=False, no_errors=False, **kargs):
        """
        Find a system via a name or keys specified in the ``**kargs``.

        :param name: The name to search for.
        :param return_list: If only the first result or all results should be returned.
        :param no_errors: Silence some errors which would raise if this turned to False.
        :param kargs: Additional key-value pairs which may help in finding the desired objects.
        :return: A single object or a list of all search results.
        """
        return self._collection_mgr.systems().find(name=name, return_list=return_list, no_errors=no_errors, **kargs)

    def find_repo(self, name: str = "", return_list=False, no_errors=False, **kargs):
        """
        Find a repository via a name or keys specified in the ``**kargs``.

        :param name: The name to search for.
        :param return_list: If only the first result or all results should be returned.
        :param no_errors: Silence some errors which would raise if this turned to False.
        :param kargs: Additional key-value pairs which may help in finding the desired objects.
        :return: A single object or a list of all search results.
        """
        return self._collection_mgr.repos().find(name=name, return_list=return_list, no_errors=no_errors, **kargs)

    def find_image(self, name: str = "", return_list=False, no_errors=False, **kargs):
        """
        Find an image via a name or keys specified in the ``**kargs``.

        :param name: The name to search for.
        :param return_list: If only the first result or all results should be returned.
        :param no_errors: Silence some errors which would raise if this turned to False.
        :param kargs: Additional key-value pairs which may help in finding the desired objects.
        :return: A single object or a list of all search results.
        """
        return self._collection_mgr.images().find(name=name, return_list=return_list, no_errors=no_errors, **kargs)

    def find_mgmtclass(self, name: str = "", return_list=False, no_errors=False, **kargs):
        """
        Find a management class via a name or keys specified in the ``**kargs``.

        :param name: The name to search for.
        :param return_list: If only the first result or all results should be returned.
        :param no_errors: Silence some errors which would raise if this turned to False.
        :param kargs: Additional key-value pairs which may help in finding the desired objects.
        :return: A single object or a list of all search results.
        """
        return self._collection_mgr.mgmtclasses().find(name=name, return_list=return_list, no_errors=no_errors, **kargs)

    def find_package(self, name: str = "", return_list=False, no_errors=False, **kargs):
        """
        Find a package via a name or keys specified in the ``**kargs``.

        :param name: The name to search for.
        :param return_list: If only the first result or all results should be returned.
        :param no_errors: Silence some errors which would raise if this turned to False.
        :param kargs: Additional key-value pairs which may help in finding the desired objects.
        :return: A single object or a list of all search results.
        """
        return self._collection_mgr.packages().find(name=name, return_list=return_list, no_errors=no_errors, **kargs)

    def find_file(self, name: str = "", return_list=False, no_errors=False, **kargs):
        """
        Find a file via a name or keys specified in the ``**kargs``.

        :param name: The name to search for.
        :param return_list: If only the first result or all results should be returned.
        :param no_errors: Silence some errors which would raise if this turned to False.
        :param kargs: Additional key-value pairs which may help in finding the desired objects.
        :return: A single object or a list of all search results.
        """
        return self._collection_mgr.files().find(name=name, return_list=return_list, no_errors=no_errors, **kargs)

    def find_menu(self, name: str = "", return_list=False, no_errors=False, **kargs):
        """
        Find a menu via a name or keys specified in the ``**kargs``.

        :param name: The name to search for.
        :param return_list: If only the first result or all results should be returned.
        :param no_errors: Silence some errors which would raise if this turned to False.
        :param kargs: Additional key-value pairs which may help in finding the desired objects.
        :return: A single object or a list of all search results.
        """
        return self._collection_mgr.menus().find(name=name, return_list=return_list, no_errors=no_errors, **kargs)

    # ==========================================================================

    def __since(self, mtime: float, collector, collapse: bool = False) -> list:
        """
        Called by get_*_since functions. This is an internal part of Cobbler.

        :param mtime: The timestamp which marks the gate if an object is included or not.
        :param collector: The list of objects to filter after mtime.
        :param collapse: Whether the object should be collapsed to a dict or not. If not the item objects are used for
                         the list.
        :return: The list of objects which are newer then the given timestamp.
        """
        results1 = collector()
        results2 = []
        for x in results1:
            if x.mtime == 0 or x.mtime >= mtime:
                if not collapse:
                    results2.append(x)
                else:
                    results2.append(x.to_dict())
        return results2

    def get_distros_since(self, mtime: float, collapse: bool = False):
        """
        Returns distros modified since a certain time (in seconds since Epoch)

        :param mtime: The timestamp which marks the gate if an object is included or not.
        :param collapse: collapse=True specifies returning a dict instead of objects.
        :return: The list of distros which are newer then the given timestamp.
        """
        return self.__since(mtime, self.distros, collapse=collapse)

    def get_profiles_since(self, mtime: float, collapse: bool = False) -> list:
        """
        Returns profiles modified since a certain time (in seconds since Epoch)

        :param mtime: The timestamp which marks the gate if an object is included or not.
        :param collapse: If True then this specifies that a list of dicts should be returned instead of a list of
                         objects.
        :return: The list of profiles which are newer then the given timestamp.
        """
        return self.__since(mtime, self.profiles, collapse=collapse)

    def get_systems_since(self, mtime: float, collapse: bool = False) -> list:
        """
        Return systems modified since a certain time (in seconds since Epoch)

        :param mtime: The timestamp which marks the gate if an object is included or not.
        :param collapse: If True then this specifies that a list of dicts should be returned instead of a list of
                         objects.
        :return: The list of systems which are newer then the given timestamp.
        """
        return self.__since(mtime, self.systems, collapse=collapse)

    def get_repos_since(self, mtime: float, collapse: bool = False) -> list:
        """
        Return repositories modified since a certain time (in seconds since Epoch)

        :param mtime: The timestamp which marks the gate if an object is included or not.
        :param collapse: If True then this specifies that a list of dicts should be returned instead of a list of
                         objects.
        :return: The list of repositories which are newer then the given timestamp.
        """
        return self.__since(mtime, self.repos, collapse=collapse)

    def get_images_since(self, mtime: float, collapse: bool = False) -> list:
        """
        Return images modified since a certain time (in seconds since Epoch)

        :param mtime: The timestamp which marks the gate if an object is included or not.
        :param collapse: If True then this specifies that a list of dicts should be returned instead of a list of
                         objects.
        :return: The list of images which are newer then the given timestamp.
        """
        return self.__since(mtime, self.images, collapse=collapse)

    def get_mgmtclasses_since(self, mtime: float, collapse: bool = False) -> list:
        """
        Return management classes modified since a certain time (in seconds since Epoch)

        :param mtime: The timestamp which marks the gate if an object is included or not.
        :param collapse: If True then this specifies that a list of dicts should be returned instead of a list of
                         objects.
        :return: The list of management classes which are newer then the given timestamp.
        """
        return self.__since(mtime, self.mgmtclasses, collapse=collapse)

    def get_packages_since(self, mtime: float, collapse: bool = False) -> list:
        """
        Return packages modified since a certain time (in seconds since Epoch)

        :param mtime: The timestamp which marks the gate if an object is included or not.
        :param collapse: If True then this specifies that a list of dicts should be returned instead of a list of
                         objects.
        :return: The list of packages which are newer then the given timestamp.
        """
        return self.__since(mtime, self.packages, collapse=collapse)

    def get_files_since(self, mtime: float, collapse: bool = False) -> list:
        """
        Return files modified since a certain time (in seconds since Epoch)

        :param mtime: The timestamp which marks the gate if an object is included or not.
        :param collapse: If True then this specifies that a list of dicts should be returned instead of a list of
                         objects.
        :return: The list of files which are newer then the given timestamp.
        """
        return self.__since(mtime, self.files, collapse=collapse)

    def get_menus_since(self, mtime: float, collapse=False) -> list:
        """
        Return files modified since a certain time (in seconds since Epoch)

        :param mtime: The timestamp which marks the gate if an object is included or not.
        :param collapse: If True then this specifies that a list of dicts should be returned instead of a list of
                         objects.
        :return: The list of files which are newer then the given timestamp.
        """
        return self.__since(mtime, self.menus, collapse=collapse)

    # ==========================================================================

    def get_signatures(self) -> dict:
        """
        This returns the local signature cache.

        :return: The dict containing all signatures.
        """
        return utils.SIGNATURE_CACHE

    def signature_update(self):
        """
        Update all signatures from the URL specified in the settings.
        """
        try:
            url = self.settings().signature_url
            dlmgr = download_manager.DownloadManager(self._collection_mgr)
            # write temp json file
            tmpfile = tempfile.NamedTemporaryFile()
            sigjson = dlmgr.urlread(url)
            tmpfile.write(sigjson.text.encode())
            tmpfile.flush()
            self.logger.debug("Successfully got file from %s" % self.settings().signature_url)
            # test the import without caching it
            try:
                utils.load_signatures(tmpfile.name, cache=False)
            except:
                self.logger.error("Downloaded signatures failed test load (tempfile = %s)", tmpfile.name)

            # rewrite the real signature file and import it for real
            f = open(self.settings().signature_path, "w")
            f.write(sigjson.text)
            f.close()

            utils.load_signatures(self.settings().signature_path)
        except:
            utils.log_exc()

    # ==========================================================================

    def dump_vars(self, obj, formatted_output: bool = False):
        """
        Dump all known variables related to that object.

        :param obj: The object for which the variables should be dumped.
        :param formatted_output: If True the values will align in one column and be pretty printed for cli example.
        :return: A dictionary with all the information which could be collected.
        """
        return obj.dump_vars(formatted_output)

    # ==========================================================================

    def auto_add_repos(self):
        """
        Import any repos this server knows about and mirror them. Run ``cobbler reposync`` to apply the changes.
        Credit: Seth Vidal.

        :raises ImportError
        """
        self.log("auto_add_repos")
        try:
            import dnf
        except:
            raise ImportError("dnf is not installed")

        base = dnf.Base()
        base.read_all_repos()
        basearch = base.conf.substitutions["basearch"]

        for repository in base.repos.iter_enabled():
            auto_name = repository.id + '-' + base.conf.releasever + '-' + basearch

            if self.find_repo(auto_name) is None:
                cobbler_repo = self.new_repo()
                cobbler_repo.name = auto_name
                cobbler_repo.breed = enums.RepoBreeds.YUM
                cobbler_repo.arch = basearch
                cobbler_repo.comment = repository.name
                baseurl = repository.baseurl
                metalink = repository.metalink
                mirrorlist = repository.mirrorlist

                if metalink is not None:
                    mirror = metalink
                    mirror_type = enums.MirrorType.METALINK
                elif mirrorlist is not None:
                    mirror = mirrorlist
                    mirror_type = enums.MirrorType.MIRRORLIST
                elif len(baseurl) > 0:
                    mirror = baseurl[0]
                    mirror_type = enums.MirrorType.BASEURL
                else:
                    mirror = ""
                    mirror_type = enums.MirrorType.NONE

                cobbler_repo.mirror = mirror
                cobbler_repo.mirror_type = mirror_type
                self.log("auto repo adding: %s" % auto_name)
                self.add_repo(cobbler_repo)
            else:
                self.log("auto repo adding: %s - exists" % auto_name)

    # ==========================================================================

    def get_repo_config_for_profile(self, obj) -> str:
        """
        Get the repository configuration for the specified profile

        :param obj: The profile to return the configuration for.
        :return: The repository configuration as a string.
        """
        return self.yumgen.get_yum_config(obj, True)

    def get_repo_config_for_system(self, obj) -> str:
        """
        Get the repository configuration for the specified system.

        :param obj: The system to return the configuration for.
        :return: The repository configuration as a string.
        """
        return self.yumgen.get_yum_config(obj, False)

    # ==========================================================================

    def get_template_file_for_profile(self, obj, path) -> str:
        """
        Get the template for the specified profile.

        :param obj: The object which is related to that template.
        :param path: The path to the template.
        :return: The template as in its string representation.
        """
        template_results = self.tftpgen.write_templates(obj, False, path)
        if path in template_results:
            return template_results[path]
        else:
            return "# template path not found for specified profile"

    def get_template_file_for_system(self, obj, path):
        """
        Get the template for the specified system.

        :param obj: The object which is related to that template.
        :param path: The path to the template.
        :return: The template as in its string representation.
        """
        template_results = self.tftpgen.write_templates(obj, False, path)
        if path in template_results:
            return template_results[path]
        else:
            return "# template path not found for specified system"

    # ==========================================================================

    def generate_ipxe(self, profile: str, image: str, system: str) -> str:
        """
        Generate the ipxe configuration files. The system wins over the profile. Profile and System win over Image.

        :param profile: The profile to return the configuration for.
        :param image: The image to return the configuration for.
        :param system: The system to return the configuration for.
        :return: The generated configuration file.
        """
        self.log("generate_ipxe")
        data = ""
        if profile is None and image is None and system is None:
            boot_menu = self.tftpgen.make_pxe_menu()
            if 'ipxe' in boot_menu:
                data = boot_menu['ipxe']
        elif system:
            data = self.tftpgen.generate_ipxe("system", system)
        elif profile:
            data = self.tftpgen.generate_ipxe("profile", profile)
        elif image:
            data = self.tftpgen.generate_ipxe("image", image)
        return data

    # ==========================================================================

    def generate_bootcfg(self, profile: str, system: str) -> str:
        """
        Generate a boot configuration. The system wins over the profile.

        :param profile: The profile to return the configuration for.
        :param system: The system to return the configuration for.
        :return: The generated configuration file.
        """
        self.log("generate_bootcfg")
        if system:
            return self.tftpgen.generate_bootcfg("system", system)
        else:
            return self.tftpgen.generate_bootcfg("profile", profile)

    # ==========================================================================

    def generate_script(self, profile: Optional[str], system: Optional[str], name: str):
        """
        Generate an autoinstall script for the specified profile or system. The system wins over the profile.

        :param profile: The profile name to generate the script for.
        :param system: The system name to generate the script for.
        :param name: The name of the script which should be generated. Must only contain alphanumeric characters, dots
                     and underscores.
        :return: The generated script or an error message.
        """
        self.log("generate_script")
        if system:
            return self.tftpgen.generate_script("system", system, name)
        else:
            return self.tftpgen.generate_script("profile", profile, name)

    # ==========================================================================

    def check(self) -> Union[None, list]:
        """
        See if all preqs for network booting are valid. This returns a list of strings containing instructions on things
        to correct. An empty list means there is nothing to correct, but that still doesn't mean there are configuration
        errors. This is mainly useful for human admins, who may, for instance, forget to properly set up their TFTP
        servers for PXE, etc.

        :return: None or a list of things to address.
        """
        self.log("check")
        action_check = check.CobblerCheck(self._collection_mgr)
        return action_check.run()

    # ==========================================================================

    def validate_autoinstall_files(self):
        """
        Validate if any of the autoinstallation files are invalid and if yes report this.

        """
        self.log("validate_autoinstall_files")
        autoinstall_mgr = autoinstall_manager.AutoInstallationManager(self._collection_mgr)
        autoinstall_mgr.validate_autoinstall_files()

    # ==========================================================================

    def sync_systems(self, systems: List[str], verbose: bool = False):
        """
        Take the values currently written to the configuration files in /etc, and /var, and build out the information
        tree found in /tftpboot. Any operations done in the API that have not been saved with serialize() will NOT be
        synchronized with this command.

        :param systems: List of specified systems that needs to be synced
        :param verbose: If the action should be just logged as needed or (if True) as much verbose as possible.
        """
        self.log("sync_systems")
        if not (systems and isinstance(systems, list) and all(isinstance(sys_name, str) for sys_name in systems)):
            raise TypeError('Systems must be a list of one or more strings.')
        sync_obj = self.get_sync(verbose=verbose)
        sync_obj.run_sync_systems(systems)

    # ==========================================================================

    def sync(self, verbose: bool = False, what: Optional[list] = None):
        """
        Take the values currently written to the configuration files in /etc, and /var, and build out the information
        tree found in /tftpboot. Any operations done in the API that have not been saved with serialize() will NOT be
        synchronized with this command.

        :param verbose: If the action should be just logged as needed or (if True) as much verbose as possible.
        :param what:   List of strings what services to sync (e.g. dhcp and/or dns). Empty list for full sync.
        """
        # Empty what: Full sync
        if not what:
            self.logger.info("syncing all")
            sync_obj = self.get_sync(verbose=verbose)
            sync_obj.run()
            return
        # Non empty what: Specific sync
        if not isinstance(what, list):
            raise TypeError("'what' needs to be of type list!")
        if 'dhcp' in what:
            self.sync_dhcp()
        if 'dns' in what:
            self.sync_dns()

    # ==========================================================================

    def sync_dns(self):
        """
        Only build out the DNS configuration.
        """
        if not self.settings().manage_dns:
            self.logger.error("manage_dns not set")
            return
        self.logger.info("sync_dns")
        dns_module = self.get_module_from_file(
            "dns",
            "module",
            "managers.bind"
        )
        dns = dns_module.get_manager(self._collection_mgr)
        dns.sync()

    # ==========================================================================

    def sync_dhcp(self):
        """
        Only build out the DHCP configuration.
        """
        if not self.settings().manage_dhcp:
            self.logger.error("manage_dhcp not set")
            return
        self.logger.info("sync_dhcp")
        dhcp_module = self.get_module_from_file(
            "dhcp",
            "module",
            "managers.isc"
        )
        dhcp = dhcp_module.get_manager(self._collection_mgr)
        dhcp.sync()

    # ==========================================================================

    def get_sync(self, verbose: bool = False):
        """
        Get a Cobbler Sync object which may be executed through the call of ``obj.run()``.

        :param verbose: If the action should be just logged as needed or (if True) as much verbose as possible.
        :return: An instance of the CobblerSync class to execute the sync with.
        """
        if not isinstance(verbose, bool):
            raise TypeError("get_sync: verbose parameter needs to be of type bool!")
        dhcp = self.get_module_from_file(
            "dhcp",
            "module",
            "managers.isc"
        ).get_manager(self._collection_mgr)
        dns = self.get_module_from_file(
            "dns",
            "module",
            "managers.bind"
        ).get_manager(self._collection_mgr)
        tftpd = self.get_module_from_file(
            "tftpd",
            "module",
            "managers.in_tftpd",
        ).get_manager(self._collection_mgr)

        return sync.CobblerSync(self._collection_mgr, dhcp=dhcp, dns=dns, tftpd=tftpd, verbose=verbose)

    # ==========================================================================

    def reposync(self, name: Optional[str] = None, tries: int = 1, nofail: bool = False):
        """
        Take the contents of ``/var/lib/cobbler/repos`` and update them -- or create the initial copy if no contents
        exist yet.

        :param name: The name of the repository to run reposync for.
        :param tries: How many tries should be executed before the action fails.
        :param nofail: If True then the action will fail, otherwise the action will just be skipped. This respects the
                       ``tries`` parameter.
        """
        self.log("reposync", [name])
        action_reposync = reposync.RepoSync(self._collection_mgr, tries=tries, nofail=nofail)
        action_reposync.run(name)

    # ==========================================================================

    def status(self, mode: str):
        """
        Get the status of the current Cobbler instance.

        :param mode: "text" or anything else. Meaning whether the output is thought for the terminal or not.
        :return: The current status of Cobbler.
        """
        statusifier = status.CobblerStatusReport(self._collection_mgr, mode)
        return statusifier.run()

    # ==========================================================================

    def import_tree(self, mirror_url: str, mirror_name: str, network_root=None, autoinstall_file=None, rsync_flags=None,
                    arch=None, breed=None, os_version=None) -> bool:
        """
        Automatically import a directory tree full of distribution files.

        :param mirror_url: Can be a string that represents a path, a user@host syntax for SSH, or an rsync:// address.
                           If mirror_url is a filesystem path and mirroring is not desired, set network_root to
                           something like "nfs://path/to/mirror_url/root"
        :param mirror_name: The name of the mirror.
        :param network_root:
        :param autoinstall_file:
        :param rsync_flags:
        :param arch:
        :param breed:
        :param os_version:
        """
        self.log("import_tree", [mirror_url, mirror_name, network_root, autoinstall_file, rsync_flags])

        # Both --path and --name are required arguments.
        if mirror_url is None or not mirror_url:
            self.log("import failed.  no --path specified")
            return False
        if mirror_name is None or not mirror_name:
            self.log("import failed.  no --name specified")
            return False

        path = os.path.normpath("%s/distro_mirror/%s" % (self.settings().webdir, mirror_name))
        if arch is not None:
            arch = arch.lower()
            if arch == "x86":
                # be consistent
                arch = "i386"
            if path.split("-")[-1] != arch:
                path += ("-%s" % arch)

        # We need to mirror (copy) the files.
        self.log("importing from a network location, running rsync to fetch the files first")

        utils.mkdir(path)

        # Prevent rsync from creating the directory name twice if we are copying via rsync.

        if not mirror_url.endswith("/"):
            mirror_url = "%s/" % mirror_url

        if mirror_url.startswith("http://") or mirror_url.startswith("https://") or mirror_url.startswith("ftp://") \
                or mirror_url.startswith("nfs://"):
            # HTTP mirrors are kind of primative. rsync is better. That's why this isn't documented in the manpage and
            # we don't support them.
            # TODO: how about adding recursive FTP as an option?
            self.log("unsupported protocol")
            return False
        else:
            # Good, we're going to use rsync.. We don't use SSH for public mirrors and local files.
            # Presence of user@host syntax means use SSH
            spacer = ""
            if not mirror_url.startswith("rsync://") and not mirror_url.startswith("/"):
                spacer = ' -e "ssh" '
            rsync_cmd = RSYNC_CMD
            if rsync_flags:
                rsync_cmd += " " + rsync_flags

            # If --available-as was specified, limit the files we pull down via rsync to just those that are critical
            # to detecting what the distro is
            if network_root is not None:
                rsync_cmd += " --include-from=/etc/cobbler/import_rsync_whitelist"

            # kick off the rsync now
            utils.run_this(rsync_cmd, (spacer, mirror_url, path))

        if network_root is not None:
            # In addition to mirroring, we're going to assume the path is available over http, ftp, and nfs, perhaps on
            # an external filer. Scanning still requires --mirror is a filesystem path, but --available-as marks the
            # network path. This allows users to point the path at a directory containing just the network boot files
            # while the rest of the distro files are available somewhere else.

            # Find the filesystem part of the path, after the server bits, as each distro URL needs to be calculated
            # relative to this.

            if not network_root.endswith("/"):
                network_root += "/"
            valid_roots = ["nfs://", "ftp://", "http://", "https://"]
            for valid_root in valid_roots:
                if network_root.startswith(valid_root):
                    break
            else:
                self.log("Network root given to --available-as must be nfs://, ftp://, http://, or https://")
                return False

            if network_root.startswith("nfs://"):
                try:
                    (a, b, rest) = network_root.split(":", 3)
                except:
                    self.log("Network root given to --available-as is missing a colon, please see the manpage example.")
                    return False

        import_module = self.get_module_by_name("managers.import_signatures") \
            .get_import_manager(self._collection_mgr)
        import_module.run(path, mirror_name, network_root, autoinstall_file, arch, breed, os_version)
        return True

    # ==========================================================================

    def acl_config(self, adduser: Optional[str] = None, addgroup: Optional[str] = None,
                   removeuser: Optional[str] = None, removegroup: Optional[str] = None):
        """
        Configures users/groups to run the Cobbler CLI as non-root.
        Pass in only one option at a time. Powers ``cobbler aclconfig``.

        :param adduser:
        :param addgroup:
        :param removeuser:
        :param removegroup:
        """
        action_acl = acl.AclConfig(self._collection_mgr)
        action_acl.run(
            adduser=adduser,
            addgroup=addgroup,
            removeuser=removeuser,
            removegroup=removegroup
        )

    # ==========================================================================

    def serialize(self):
        """
        Save the cobbler_collections to disk.
        Cobbler internal use only.
        """
        self._collection_mgr.serialize()

    def deserialize(self):
        """
        Load cobbler_collections from disk.
        Cobbler internal use only.
        """
        return self._collection_mgr.deserialize()

    # ==========================================================================

    def get_module_by_name(self, module_name: str):
        """
        Returns a loaded Cobbler module named 'name', if one exists, else None.
        Cobbler internal use only.

        :param module_name:
        :return:
        """
        return module_loader.get_module_by_name(module_name)

    def get_module_from_file(self, section: str, name: str, fallback: Optional[str] = None):
        """
        Looks in ``/etc/cobbler/modules.conf`` for a section called 'section' and a key called 'name', and then returns
        the module that corresponds to the value of that key.
        Cobbler internal use only.

        :param section:
        :param name:
        :param fallback:
        :return:
        """
        return module_loader.get_module_from_file(section, name, fallback)

    def get_module_name_from_file(self, section: str, name: str, fallback: Optional[str] = None):
        """
        Looks up a module the same as ``get_module_from_file`` but returns the module name rather than the module
        itself.

        :param section:
        :param name:
        :param fallback:
        :return:
        """
        return module_loader.get_module_name(section, name, fallback)

    def get_modules_in_category(self, category: str):
        """
        Returns all modules in a given category, for instance "serializer", or "cli".
        Cobbler internal use only.

        :param category: The category to check.
        :return: The list of modules.
        """
        return module_loader.get_modules_in_category(category)

    # ==========================================================================

    def authenticate(self, user: str, password: str):
        """
        (Remote) access control. This depends on the chosen authentication module.
        Cobbler internal use only.

        :param user: The username to check for authentication.
        :param password: The password to check for authentication.
        :return: Whether the action succeeded or not.
        """
        rc = self.authn.authenticate(self, user, password)
        self.log("authenticate", [user, rc])
        return rc

    def authorize(self, user: str, resource: str, arg1=None, arg2=None):
        """
        (Remote) access control. This depends on the chosen authorization module.
        Cobbler internal use only.

        :param user: The username to check for authorization.
        :param resource: The type of resource which should be checked for access from the supplied user.
        :param arg1: The actual resource to check for authorization.
        :param arg2: Not known what this parameter does exactly.
        :return: The return code of the action.
        """
        rc = self.authz.authorize(self, user, resource, arg1, arg2)
        self.log("authorize", [user, resource, arg1, arg2, rc], debug=True)
        return rc

    # ==========================================================================

    def build_iso(self, iso=None, profiles=None, systems=None, buildisodir=None, distro=None, standalone=None,
                  airgapped=None, source=None, exclude_dns=None, xorrisofs_opts=None):
        """
        Build an iso image which may be network bootable or not.

        :param iso:
        :param profiles:
        :param systems:
        :param buildisodir:
        :param distro:
        :param standalone:
        :param airgapped:
        :param source:
        :param exclude_dns:
        :param xorrisofs_opts:
        """
        builder = buildiso.BuildIso(self._collection_mgr)
        builder.run(
            iso=iso, profiles=profiles, systems=systems, buildisodir=buildisodir, distro=distro, standalone=standalone,
            airgapped=airgapped, source=source, exclude_dns=exclude_dns, xorrisofs_opts=xorrisofs_opts
        )

    # ==========================================================================

    def hardlink(self):
        """
        Hardlink all files where this is possible to improve performance.

        :return: The return code of the subprocess call which actually hardlinks the files.
        """
        linker = hardlink.HardLinker(api=self)
        return linker.run()

    # ==========================================================================

    def replicate(self, cobbler_master: Optional[str] = None, port: str = "80", distro_patterns: str = "",
                  profile_patterns: str = "", system_patterns: str = "", repo_patterns: str = "",
                  image_patterns: str = "", mgmtclass_patterns=None, package_patterns=None, file_patterns: bool = False,
                  prune: bool = False, omit_data: bool = False, sync_all: bool = False, use_ssl: bool = False):
        """
        Pull down data/configs from a remote Cobbler server that is a master to this server.

        :param cobbler_master: The hostname/URL of the other Cobbler server
        :param port: The port to use for the replication task.
        :param distro_patterns: The pattern of distros which should be synced.
        :param profile_patterns: The pattern of profiles which should be synced.
        :param system_patterns: The pattern of systems which should be synced.
        :param repo_patterns: The pattern of repositories which should be synced.
        :param image_patterns: The pattern of images which should be synced.
        :param mgmtclass_patterns: The pattern of management classes which should be synced.
        :param package_patterns: The pattern of packages which should be synced.
        :param file_patterns: The pattern of files which should be synced.
        :param prune: Whether the object not on the master should be removed or not.
        :param omit_data: If the data downloaded by the current Cobbler server should be rsynced to the destination
                          server.
        :param sync_all: This parameter behaves similarly to a dry run argument. If True then everything will executed,
                         if False then only some things are synced.
        :param use_ssl: Whether SSL should be used (True) or not (False).
        """
        replicator = replicate.Replicate(self._collection_mgr)
        return replicator.run(
            cobbler_master=cobbler_master, port=port, distro_patterns=distro_patterns,
            profile_patterns=profile_patterns, system_patterns=system_patterns, repo_patterns=repo_patterns,
            image_patterns=image_patterns, mgmtclass_patterns=mgmtclass_patterns, package_patterns=package_patterns,
            file_patterns=file_patterns, prune=prune, omit_data=omit_data, sync_all=sync_all, use_ssl=use_ssl
        )

    # ==========================================================================

    def report(self, report_what: Optional[str] = None, report_name: Optional[str] = None,
               report_type: Optional[str] = None, report_fields: Optional[str] = None,
               report_noheaders: Optional[bool] = None):
        """
        Report functionality for Cobbler.

        :param report_what: The object type that should be reported.
        :param report_name: The name of the object which should be possibly reported.
        :param report_type: May be either "text", "csv", "mediawiki", "trac" or "doku".
        :param report_fields: Specify "all" or the fields you want to be reported.
        :param report_noheaders: If the column headers should be included in the output or not.
        """
        reporter = report.Report(self._collection_mgr)
        return reporter.run(report_what=report_what, report_name=report_name, report_type=report_type,
                            report_fields=report_fields, report_noheaders=report_noheaders)

    # ==========================================================================

    def power_system(self, system: str, power_operation: str, user: Optional[str] = None,
                     password: Optional[str] = None):
        """
        Power on / power off / get power status /reboot a system.

        :param system: Cobbler system
        :param power_operation: power operation. Valid values: on, off, reboot, status
        :param user: power management user
        :param password: power management password
        :return: bool if operation was successful
        """

        if power_operation == "on":
            self.power_mgr.power_on(system, user=user, password=password)
        elif power_operation == "off":
            self.power_mgr.power_off(system, user=user, password=password)
        elif power_operation == "status":
            return self.power_mgr.get_power_status(system, user=user, password=password)
        elif power_operation == "reboot":
            self.power_mgr.reboot(system, user=user, password=password)
        else:
            utils.die("invalid power operation '%s', expected on/off/status/reboot" % power_operation)
        return None

    # ==========================================================================

    def clear_logs(self, system):
        """
        Clears console and anamon logs for system

        :param system: The system to clear logs of.
        """
        log.LogTool(self._collection_mgr, system, self).clear()

    # ==========================================================================

    def get_valid_obj_boot_loaders(self, obj) -> List[str]:
        """
        Return the list of valid boot loaders for the object

        :param obj: The object for which the boot loaders should be looked up.
        :return: Get a list of all valid boot loaders.
        """
        return obj.supported_boot_loaders

    # ==========================================================================

    def create_grub_images(self):
        """
        Create the GRUB installer images via this API call. It utilizes ``grub2-mkimage`` behind the curtain.
        """
        action = grubimage.GrubImage(self)
        action.run()
